#*******************************************************************************************************************************
#
# This file contains the Vivado Tcl commands to create a Model 37X-YYY SigFPGA project.
# Note:	37X-YYY is the specific part number of the product you ordered from Red Rapids (e.g. Model 377-403).
#
# Instructions to execute from Vivado GUI mode:
# Enter the following three commands from the Vivado Tcl console:
#   set argv [ list 37X-YYY ]
#	set argc 1
#	source project.tcl
#
# Instructions to execute from Vivado batch mode:
# Enter the following command from an OS command shell:
#   vivado -mode batch -notrace -nojournal -nolog -source project.tcl -tclargs 37X-YYY
# Note: The "vivado" executable must be in your search path, otherwise provide the complete pathname.
#
#*******************************************************************************************************************************

# Break the part number (XXX-YYY) into a model number (XXX) and build option code (YYY).
set modelnumber [string range [lindex $argv 0] 0 2]
set buildoption [string range [lindex $argv 0] 4 6]

# Set variables based on the part number.
switch -exact $modelnumber {
	371 { switch -exact $buildoption {
		001 { set physical "XMC"; set device "xc7k410tfbg676-2"; set sram "FF" }
		002 { set physical "XMC"; set device "xc7k410tfbg676-2"; set sram "NO" }
		100 { set physical "XMC"; set device "xc7k160tfbg676-2"; set sram "NO" }
		400 { set physical "PCI"; set device "xc7k410tfbg676-2"; set sram "SS" }
		default { puts "ERROR: Build option $buildoption is not available for model number $modelnumber"; exit }
		}
	}
	372 { switch -exact $buildoption {
		002 { set physical "XMC"; set device "xc7k325tfbg676-2"; set sram "SS" }
		500 { set physical "XMC"; set device "xc7k325tfbg676-2"; set sram "NO" }
		501 { set physical "XMC"; set device "xc7k325tfbg676-2"; set sram "NO" }
		default { puts "ERROR: Build option $buildoption is not available for model number $modelnumber"; exit }
		}
	}
	373 { switch -exact $buildoption {
		100 { set physical "XMC"; set device "xc7k160tfbg676-2"; set sram "NO" }
		default { puts "ERROR: Build option $buildoption is not available for model number $modelnumber"; exit }
		}
	}
	376 { switch -exact $buildoption {
		001 { set physical "XMC"; set device "xc7k410tfbg676-2"; set sram "FF" }
		003 { set physical "XMC"; set device "xc7k410tfbg676-2"; set sram "NO" }
		101 { set physical "XMC"; set device "xc7k410tfbg676-2"; set sram "FF" }
		default { puts "ERROR: Build option $buildoption is not available for model number $modelnumber"; exit }
		}
	}
	377 { switch -exact $buildoption {
		100 { set physical "XMC"; set device "xc7k325tfbg676-2"; set sram "SS" }
		403 { set physical "PCI"; set device "xc7k160tfbg676-2"; set sram "FF" }
		405 { set physical "PCI"; set device "xc7k410tfbg676-2"; set sram "FF" }
		default { puts "ERROR: Build option $buildoption is not available for model number $modelnumber"; exit }
		}
	}
	default { puts "ERROR: Model number $modelnumber is invalid"; exit }
}

# Print the results of the part number decoder.
puts "Model Number: $modelnumber"
puts "Build Option: $buildoption"
puts "Physical: $physical"
puts "Device: $device"
puts "SRAM: $sram"


# Create a Managed IP project.
create_project managed_ip_project ../ip_sources/vivado/$device/managed_ip_project -part $device -ip -force

# Set project properties.
set obj [get_projects managed_ip_project]
set_property "simulator_language" "Mixed" $obj
set_property "target_language" "VHDL" $obj

# Set 'sources_1' fileset object.
set obj [get_filesets sources_1]
set files [list \
 "[file normalize "./../ip_sources/vivado/$device/dma_on_demand_memory_v1/dma_on_demand_memory.xci"]"\
 "[file normalize "./../ip_sources/vivado/$device/fifo_channel_1kx64_v1/fifo_channel_1kx64.xci"]"\
 "[file normalize "./../ip_sources/vivado/$device/fifo_dma_1kx64_v1/fifo_dma_1kx64.xci"]"\
 "[file normalize "./../ip_sources/vivado/$device/fifo_flash_512x32_v0/fifo_flash_512x32.xci"]"\
 "[file normalize "./../ip_sources/vivado/$device/pcie_ep_x8_gen2_axi128_250mhz_v1/pcie_endpoint.xci"]"\
]
# Include SRAM source files on models equipped with memory.
if {$sram != "NO"} {
  set files [concat $files \
   "[file normalize "./../ip_sources/vivado/$device/fifo_qdr_512x36_v0/fifo_qdr_512x36.xci"]"\
  ]
}
add_files -norecurse -fileset $obj $files

# Create a list of IP cores that need to have output products generated.
set ip {}

generate_target all [get_files  ./../ip_sources/vivado/$device/dma_on_demand_memory_v1/dma_on_demand_memory.xci]
export_ip_user_files -of_objects [get_files ./../ip_sources/vivado/$device/dma_on_demand_memory_v1/dma_on_demand_memory.xci] -no_script -force -quiet
set ip [concat $ip [create_ip_run [get_files -of_objects [get_fileset sources_1] ./../ip_sources/vivado/$device/dma_on_demand_memory_v1/dma_on_demand_memory.xci]]]

generate_target all [get_files  ./../ip_sources/vivado/$device/fifo_channel_1kx64_v1/fifo_channel_1kx64.xci]
export_ip_user_files -of_objects [get_files ./../ip_sources/vivado/$device/fifo_channel_1kx64_v1/fifo_channel_1kx64.xci] -no_script -force -quiet
set ip [concat $ip [create_ip_run [get_files -of_objects [get_fileset sources_1] ./../ip_sources/vivado/$device/fifo_channel_1kx64_v1/fifo_channel_1kx64.xci]]]

generate_target all [get_files  ./../ip_sources/vivado/$device/fifo_dma_1kx64_v1/fifo_dma_1kx64.xci]
export_ip_user_files -of_objects [get_files ./../ip_sources/vivado/$device/fifo_dma_1kx64_v1/fifo_dma_1kx64.xci] -no_script -force -quiet
set ip [concat $ip [create_ip_run [get_files -of_objects [get_fileset sources_1] ./../ip_sources/vivado/$device/fifo_dma_1kx64_v1/fifo_dma_1kx64.xci]]]

generate_target all [get_files  ./../ip_sources/vivado/$device/fifo_flash_512x32_v0/fifo_flash_512x32.xci]
export_ip_user_files -of_objects [get_files ./../ip_sources/vivado/$device/fifo_flash_512x32_v0/fifo_flash_512x32.xci] -no_script -force -quiet
set ip [concat $ip [create_ip_run [get_files -of_objects [get_fileset sources_1] ./../ip_sources/vivado/$device/fifo_flash_512x32_v0/fifo_flash_512x32.xci]]]

generate_target all [get_files  ./../ip_sources/vivado/$device/pcie_ep_x8_gen2_axi128_250mhz_v1/pcie_endpoint.xci]
export_ip_user_files -of_objects [get_files ./../ip_sources/vivado/$device/pcie_ep_x8_gen2_axi128_250mhz_v1/pcie_endpoint.xci] -no_script -force -quiet
set ip [concat $ip [create_ip_run [get_files -of_objects [get_fileset sources_1] ./../ip_sources/vivado/$device/pcie_ep_x8_gen2_axi128_250mhz_v1/pcie_endpoint.xci]]]

if {$sram != "NO"} {
  generate_target all [get_files  ./../ip_sources/vivado/$device/fifo_qdr_512x36_v0/fifo_qdr_512x36.xci]
  export_ip_user_files -of_objects [get_files ./../ip_sources/vivado/$device/fifo_qdr_512x36_v0/fifo_qdr_512x36.xci] -no_script -force -quiet
  set ip [concat $ip [create_ip_run [get_files -of_objects [get_fileset sources_1] ./../ip_sources/vivado/$device/fifo_qdr_512x36_v0/fifo_qdr_512x36.xci]]]
}

# Launch synthesis on all cores that need output products.
if {[llength $ip] != 0} {
  launch_run -jobs 12 $ip
  foreach ip_synth $ip {
    if {[lsearch -exact $ip $ip_synth] != -1} {wait_on_run $ip_synth}
  }
}

# Close the Managed IP project.
close_project

# Create reference design project.
create_project SigFPGA_Top ./Model_$modelnumber-$buildoption -part $device -force

# Set the directory path for the new project.
set proj_dir [get_property directory [current_project]]

# Set project properties.
set obj [get_projects SigFPGA_Top]
set_property "default_lib" "xil_defaultlib" $obj
set_property "part" $device $obj
set_property "sim.ip.auto_export_scripts" "1" $obj
set_property "simulator_language" "Mixed" $obj
set_property "target_language" "VHDL" $obj
set_property "xpm_libraries" "XPM_CDC XPM_MEMORY" $obj

# Create 'sources_1' fileset (if not found).
if {[string equal [get_filesets -quiet sources_1] ""]} {
  create_fileset -srcset sources_1
}

# Set 'sources_1' fileset object.
set obj [get_filesets sources_1]
# Include source files that are common to all model numbers.
set files [list \
 "[file normalize "./../../vhdl_models/clock_crossing_v0.vhd"]"\
 "[file normalize "./../../vhdl_models/transceiver_package_v4.vhd"]"\
 "[file normalize "./../../vhdl_models/clock_lock_v4.vhd"]"\
 "[file normalize "./../../vhdl_models/tx_flow_control_v1.vhd"]"\
 "[file normalize "./../../vhdl_models/synchronizer_v0.vhd"]"\
 "[file normalize "./../../vhdl_models/rx_flow_control_v2.vhd"]"\
 "[file normalize "./../../vhdl_models/pci_transaction_decoder_v0.vhd"]"\
 "[file normalize "./../../vhdl_models/interrupt_timer_v0.vhd"]"\
 "[file normalize "./../../vhdl_models/dma_write_fifo_v1.vhd"]"\
 "[file normalize "./../../vhdl_models/dma_read_fifo_v1.vhd"]"\
 "[file normalize "./../../vhdl_models/dma_package_v0.vhd"]"\
 "[file normalize "./../../vhdl_models/dma_on_demand_v0.vhd"]"\
 "[file normalize "./../../vhdl_models/pci_performance_v5.vhd"]"\
 "[file normalize "./../../vhdl_models/pci_express_v1.vhd"]"\
 "[file normalize "./../../vhdl_models/bpi_flash_v2.vhd"]"\
 "[file normalize "./../../vhdl_models/bar0_decoder_v6.vhd"]"\
 "[file normalize "./../../vhdl_models/adc_interface_v4.vhd"]"\
 "[file normalize "./../../vhdl_models/serial_bus_v3.vhd"]"\
 "[file normalize "./../../vhdl_models/dac_interface_v1.vhd"]"\
 "[file normalize "./../../vhdl_models/clock_route_v2.vhd"]"\
 "[file normalize "./../../vhdl_models/interrupt_controller_v4.vhd"]"\
 "[file normalize "./../../vhdl_models/transceiver_v9.vhd"]"\
 "[file normalize "./../../vhdl_models/bar2_channel_v1.vhd"]"\
 "[file normalize "./../../vhdl_models/bar2_common_v6.vhd"]"\
 "[file normalize "./../../vhdl_models/rx_datapath_v2.vhd"]"\
 "[file normalize "./../../vhdl_models/tx_datapath_v1.vhd"]"\
 "[file normalize "./../../vhdl_models/bar2_decoder_v4.vhd"]"\
 "[file normalize "./../top_sources/M$modelnumber/$sram\_SRAM/model_specific_package_v9.vhd"]"\
 "[file normalize "./../top_sources/M$modelnumber/$sram\_SRAM/SigFPGA_top_v7.vhd"]"\
 "[file normalize "./../ip_sources/plda/ezdma2_axi128_250mhz_v0/plda_ezdma2.edn"]"\
 "[file normalize "./../ip_sources/vivado/$device/dma_on_demand_memory_v1/dma_on_demand_memory.xci"]"\
 "[file normalize "./../ip_sources/vivado/$device/fifo_channel_1kx64_v1/fifo_channel_1kx64.xci"]"\
 "[file normalize "./../ip_sources/vivado/$device/fifo_dma_1kx64_v1/fifo_dma_1kx64.xci"]"\
 "[file normalize "./../ip_sources/vivado/$device/fifo_flash_512x32_v0/fifo_flash_512x32.xci"]"\
 "[file normalize "./../ip_sources/vivado/$device/pcie_ep_x8_gen2_axi128_250mhz_v1/pcie_endpoint.xci"]"\
]
# Include SRAM source files on models equipped with memory.
if {$sram != "NO"} {
  set files [concat $files \
   "[file normalize "./../../vhdl_models/qdr_dll_v1.vhd"]"\
   "[file normalize "./../../vhdl_models/qdr_sram_v3.vhd"]"\
   "[file normalize "./../../vhdl_models/qdr_demo_v1.vhd"]"\
   "[file normalize "./../ip_sources/vivado/$device/fifo_qdr_512x36_v0/fifo_qdr_512x36.xci"]"\
  ]
}
add_files -norecurse -fileset $obj $files

# Create 'constrs_1' fileset (if not found)
if {[string equal [get_filesets -quiet constrs_1] ""]} {
  create_fileset -constrset constrs_1
}

# Set 'constrs_1' fileset object
set obj [get_filesets constrs_1]
# Include constraint files that are common to all model numbers.
set files [list \
 "[file normalize "./../constraints/target.xdc"]"\
 "[file normalize "./../top_sources/M$modelnumber/model_specific_constraints_v1.xdc"]"\
 "[file normalize "./../constraints/SigFPGA_$physical\_v4.xdc"]"\
]
# Include SRAM constraint files on models equipped with memory.
if {$sram == "FF"} {
  set files [concat $files \
   "[file normalize "./../constraints/dual_qdr_ff_sram_v3.xdc"]"\
  ]
}
if {$sram == "SS"} {
  set files [concat $files \
   "[file normalize "./../constraints/dual_qdr_ss_sram_v3.xdc"]"\
  ]
}
add_files -norecurse -fileset $obj $files

# Set 'constrs_1' fileset properties
set obj [get_filesets constrs_1]
set_property "target_constrs_file" "[file normalize "./../constraints/target.xdc"]" $obj

close_project

puts "INFO: Project SigFPGA_Top created."
