--*******************************************************************************************************************************
--
-- Description: This module performs all PCI Express write/read operations to base address register (BAR) zero.
--				The AXI_WIDTH constant determines whether slave_dataout is 64-bit or 128-bit aligned.
--					
-- Rev History: 08-04-2016:	Modified the bus range of several signals to eliminate synthesis warnings.
-- 				07-14-2015:	Changed the function of the three command and status signals to eliminate the clock_init model.
-- 				04-17-2015:	Added functionality to program flash through the PCI interface.
--				12-08-2014:	Moved GPIO and USER control bits (direction/tristate) to BAR0 registers.
--				08-20-2014:	Corrected 64-bit write decoder to address x0020, expanded the serial_config and error_* registers.
-- 				04-16-2014:	Decoupled DMA channel count from TX/RX channel count to support eight DMA channels.
-- 				09-04-2013:	Inital release.
--
--*******************************************************************************************************************************

library ieee;
use ieee.std_logic_1164.all;
use IEEE.numeric_std.all;

use work.model_specific_package.all;
use work.dma_package.all;

entity bar0_decoder is
port (
    -- Reset and clock.
    rst	                : in std_logic;									-- Active high asynchronous reset.
    clk                 : in std_logic;									-- Bus clock.

    -- PCI Express write transaction interface.
    slv_addr            : in std_logic_vector(13 downto 3);				-- Refer to the PLDA EZDMA reference manual.
    slv_dataout         : in std_logic_vector(AXI_WIDTH-1 downto 0);	-- Refer to the PLDA EZDMA reference manual.
    slv_bytevalid       : in std_logic_vector((AXI_WIDTH/8)-1 downto 0);-- Refer to the PLDA EZDMA reference manual.
    bar0_write          : in std_logic;									-- Logic one indicates an active slave write to BAR2 address space.

    -- PCI Express read transaction interface.
	bar0_read			: in std_logic;									-- Logic one indicates that the current PCI read request is directed at BAR0.
    bar0_rddata		    : out std_logic_vector(AXI_WIDTH-1 downto 0);	-- Data return from read operation on BAR0.
    dma_rd              : in std_logic;									-- Refer to the PLDA EZDMA reference manual.
    dma_rdaddr          : in  std_logic_vector(13 downto 3);			-- Refer to the PLDA EZDMA reference manual.
	dma_rdchannel 		: in  std_logic;								-- Refer to the PLDA EZDMA reference manual.

    -- DMA command and status.
	dma_current_page	: in DMA_TYPE_10;								-- Page number of the active DMA buffer.
	dma_current_burst	: in DMA_TYPE_16;								-- Number of the active DMA burst within a page.
	dma_page_address	: out DMA_TYPE_64;								-- Page address stored in the BAR0 memory.
	dma_descriptor		: out DMA_TYPE_32;								-- Page descriptor stored in the BAR0 memory.
	dma_engine			: in DMA_3;										-- Status of DMA engine busy (2), go (1), and enable (0) bits.

    -- Consolidated interrupt mask and status.
	global_int_mask 	: out std_logic;								-- Global interrupt mask to enable/disable all interrupts.
	global_status		: in std_logic_vector(4 downto 0);				-- Consolidated interrupt status bus.  
	int_timeout			: out std_logic_vector(27 downto 0);			-- Maximum time to hold an interrupt request active.
	int_response		: in std_logic_vector(27 downto 0);				-- Time taken to service the last interrupt or timeout value.

	-- Serial peripheral interface bus.
	serial_config    	: out std_logic_vector(9 downto 0);				-- Serial bus command word.
	serial_payload	    : out std_logic_vector(7 downto 0);				-- Serial bus payload (instruction + data).
	serial_start		: out std_logic;								-- Logic one initiates a serial bus transaction.
	serial_busy			: in std_logic;									-- Logic one indicates that the process is active.
	serial_shift		: in std_logic;									-- Logic one indicates that the register needs to perform a serial shift.
	serial_read			: in std_logic;									-- Registered duplicate of sdo.      

	-- GPIO and USER control bits (direction/tristate).
	gpio_direction		: out std_logic_vector(GPIO_COUNT downto 1);	-- Select GPIO direction as input (0) or output (1).
	gpio_tristate		: out std_logic;								-- Tristate all GPIO pins.
	user_direction		: out std_logic_vector(USER_COUNT downto 1);	-- Select USER direction as input (0) or output (1).
	user_tristate		: out std_logic;								-- Tristate all USER pins.

	-- BPI configuration flash interface bus.
	bpi_cmd_bus			: out std_logic_vector(32 downto 0);			-- Command bus that carries instructions and data to program the BPI flash.
	bpi_cmd_full		: in std_logic; 								-- Logic one indicates the BPI command FIFO cannot accept another burst.
	bpi_reply_bus		: in std_logic_vector(32 downto 0); 			-- Reply bus that returns data and errors from the BPI flash.
	bpi_reply_rd		: out std_logic; 								-- Logic one pops the BPI reply FIFO to the next value.

	-- Command and status.
	clk_lck_start		: out std_logic;								-- Initiates the clock lock sequence that sets IODELAY values for data inputs.
	clk_lck_busy		: in std_logic;									-- Logic one indicates that the clock lock process is active.
    software_reset		: out std_logic;								-- Active high reset tied to a register for software control.

    -- Error status and monitoring.
	error_monitor 		: in std_logic_vector(13 downto 0);				-- Bus used to monitor FIFO overflow or underlow errors.
	error_status 		: out std_logic_vector(13 downto 0);			-- Bus used to convey error status since last queried.
	error_int_mask 		: out std_logic_vector(13 downto 0);			-- Mask bits to suppress interrupts based on error status updates.
	bar0_address_error 	: out std_logic_vector(1 downto 0)				-- Bus used to convey a BAR0 PCI read or write error.
);
end bar0_decoder;

architecture structural of bar0_decoder is

	signal bytevalid				: std_logic_vector(1 downto 0);		-- Logic one indicates that all PCI byte valids are active.
	signal error_status_reg			: std_logic_vector(13 downto 0);	-- Bus used to monitor error status since last cleared.                                        	
	signal error_status_read		: std_logic_vector(13 downto 0);	-- Bus used to store error status before a clear so it can be read.                                        	
	signal error_int_mask_reg		: std_logic_vector(13 downto 0);	-- Mask bits to suppress interrupts based on fault status updates.
    signal serial_config_reg    	: std_logic_vector(9 downto 0);		-- Serial port configuration settings.
    signal serial_payload_reg    	: std_logic_vector(63 downto 0);	-- Serial port payload bits.
	signal gpio_config_reg			: std_logic_vector(GPIO_COUNT downto 0);	-- Select GPIO direction and tristate (bit 0).
	signal user_config_reg			: std_logic_vector(USER_COUNT downto 0);	-- Select USER direction and tristate (bit 0).
    signal bpi_cmd_bus_reg    		: std_logic_vector(32 downto 0);	-- Command bus that carries instructions and data to program the BPI flash.
    signal global_int_mask_reg 		: std_logic;						-- Global interrupt mask to enable/disable all interrupts.
    signal software_reset_reg    	: std_logic;						-- Register to hold a soft reset active for one clock period.
	signal software_reset_stretch	: unsigned(3 downto 0);				-- Counter to stretch the software reset.
    signal int_timeout_reg    		: std_logic_vector(27 downto 0);	-- Maximum time to hold an interrupt request active.
	signal dma_bram_bytevalid		: FIFO_BYTE;						-- Logic one indicates that all PCI byte valids are active.
    signal dma_bram_data			: FIFO_DATA;						-- DMA memory write data bus with lane steering.
    signal dma_bram_select_reg    	: std_logic_vector(15 downto 0);	-- Index to DMA page addresses and descriptors.
    signal dma_bram_we    			: std_logic;						-- DMA page address and descriptor memory write qualifier.
    signal dma_bram_page_we    		: std_logic_vector(4 downto 1);		-- DMA page address memory write enable.
    signal dma_page_data			: DMA_TYPE_64;						-- DMA page address memory data bus.
    signal dma_bram_desc_we    		: std_logic_vector(4 downto 1);		-- DMA descriptor memory write enable.
    signal dma_desc_data			: DMA_TYPE_32;						-- DMA descriptor memory data bus.
	signal dma_current_page_local	: DMA_10;							-- Page number of the active DMA buffer.
	signal dma_current_burst_local	: DMA_16;							-- Number of the active DMA burst within a page.
	signal clk_lck_stretch			: unsigned(2 downto 0);				-- Counter to stretch the clock initialization enable.
	signal clk_lck_busy_reg			: std_logic;						-- Register used to hold the clock lock busy status until cleared by a read.

begin

--***************************************************** WRITE TRANSACTIONS *****************************************************

	-- Create an active high enable for write operations to the upper 32 register bits.
	bytevalid(1) <= (slv_bytevalid(7) and slv_bytevalid(6) and slv_bytevalid(5) and slv_bytevalid(4)) when slv_addr(3) = '0'
		else (slv_bytevalid((AXI_WIDTH/8)-1) and slv_bytevalid((AXI_WIDTH/8)-2) and slv_bytevalid((AXI_WIDTH/8)-3) and slv_bytevalid((AXI_WIDTH/8)-4));
	-- Create an active high enable for write operations to the lower 32 register bits.
	bytevalid(0) <= slv_bytevalid(3) and slv_bytevalid(2) and slv_bytevalid(1) and slv_bytevalid(0) when slv_addr(3) = '0'
		else slv_bytevalid((AXI_WIDTH/8)-5) and slv_bytevalid((AXI_WIDTH/8)-6) and slv_bytevalid((AXI_WIDTH/8)-7) and slv_bytevalid((AXI_WIDTH/8)-8);
        
    process (clk,rst)
    begin
        if rst='1' then
			software_reset_reg <= '0';
            error_status_reg(13 downto 12) <= (others => '0');
            error_status_reg(ADC_CLK_SOURCES+7 downto 0) <= (others => '0');
            error_status_read(13 downto 12) <= (others => '0');
            error_status_read(ADC_CLK_SOURCES+7 downto 0) <= (others => '0');
            error_int_mask_reg <= (others => '0');
            serial_payload_reg <= (others => '0');
            serial_config_reg <= (others => '0');
            gpio_config_reg <= (0 => '1', others => '0');
            user_config_reg <= (0 => '1', others => '0');
            bpi_cmd_bus_reg <= (others => '0');
            global_int_mask_reg <= '0';
			int_timeout_reg <= (others => '1');
			dma_bram_select_reg <= (others => '0');
			bar0_address_error(0) <= '0';
			serial_start <= '0';
			clk_lck_stretch <= (others => '0');
			
        elsif rising_edge(clk) then
        
            -- Target write decoder that loads slave data into the selected register for a target write completion.
            if bar0_write = '1' then
                case "00" & slv_addr(13 downto 3) & "000" is

                	-- Software commanded reset equivalent to power on reset.
					when x"0000" =>
						if bytevalid(0) = '1' then
							-- Start the software reset counter.
							software_reset_reg <= '1';
						end if;
						
                	-- Clock initialization command.
                    when x"0008" =>
						if bytevalid(0) = '1' then
							-- Issue a single period enable strobe.
							clk_lck_stretch <= "001";
						end if;
						
					-- Error status update command defined outside of this case statement.
                    when x"0010" =>
                    	null;
					
					-- Error interrupt mask register settings.
                    when x"0018" =>
						if bytevalid(0) = '1' then error_int_mask_reg <= slv_dataout((AXI_WIDTH-64)+13 downto (AXI_WIDTH-64)); end if;
						
					-- The serial bus payload register is defined outside of this case statement.
                    when x"0020" =>
                    	null;
					
					-- Serial bus configuration register.
                    when x"0030" =>
						if bytevalid(0) = '1' then serial_config_reg <= slv_dataout(9 downto 0); end if;

 					-- Serial bus transaction command.
 					when x"0038" =>
                        if bytevalid(0) = '1' then
							-- Issue a single period enable strobe.
							serial_start <= '1';
                        end if;

					-- BPI interface control.
					when x"0040" =>
                        if bytevalid(0) = '1' then
							-- Issue a single period read strobe to the reply FIFO.
							bpi_reply_rd <= '1';
                        end if;

					-- BPI command bus.
					when x"0048" =>
                        if bytevalid(0) = '1' then
							-- Load the next instruction or data into the command FIFO.
                        	bpi_cmd_bus_reg(31 downto 0) <= slv_dataout((AXI_WIDTH-64)+31 downto (AXI_WIDTH-64));
							-- Issue a single period write strobe to the command FIFO.
							bpi_cmd_bus_reg(32) <= '1';
                        end if;

					-- GPIO control register.
                    when x"0050" =>
						if bytevalid(0) = '1' then gpio_config_reg <= slv_dataout(GPIO_COUNT downto 0); end if;

					-- USER control register.
                    when x"0058" =>
						if bytevalid(1) = '1' then user_config_reg(USER_COUNT downto 32) <= slv_dataout((AXI_WIDTH-64)+USER_COUNT downto (AXI_WIDTH-64)+32); end if;
						if bytevalid(0) = '1' then user_config_reg(31 downto 0) <= slv_dataout((AXI_WIDTH-64)+31 downto (AXI_WIDTH-64)); end if;
						
                	--  x"0100" is a read only register.
					
					-- Global interrupt mask register.
					when x"0108" =>
						if bytevalid(0) = '1' then
							global_int_mask_reg <= slv_dataout(AXI_WIDTH-64);
						end if;
						
                	--  x"0110" Read only register.
                	
					-- Interrupt timeout value.
					when x"0118" =>
						if bytevalid(0) = '1' then
							int_timeout_reg <= slv_dataout((AXI_WIDTH-64)+27 downto (AXI_WIDTH-64));
						end if;
						
                	--  x"1F00" Read only register.
                	
                	--  x"1F08" Read only register.
                	
					-- Index into the DMA page address and descriptor memories.
					when x"1FF8" =>
						if bytevalid(0) = '1' then dma_bram_select_reg <= slv_dataout((AXI_WIDTH-64)+15 downto (AXI_WIDTH-64)); end if;
						
					-- Return an error if the address is not valid.
                    when others  =>
						if slv_addr(13) = '0' then bar0_address_error(0) <= '1'; end if;
						
                end case;                                                                                                          
			else
				serial_start <= '0';
				bpi_reply_rd <= '0';
				bpi_cmd_bus_reg(32) <= '0';
				bar0_address_error(0) <= '0';
				if clk_lck_stretch /= "000" then
					clk_lck_stretch <= clk_lck_stretch + "001";
				end if;
            end if;
            
            -- Decode the write transaction to address x"0010" that will simultaneously capture and clear the fault status bits.
			if bar0_write = '1' and bytevalid(0) = '1' and ("00" & slv_addr(13 downto 3) & "000" = x"0010") then
				-- Capture the error status so that it can be read by software.
				error_status_read(13 downto 12) <= error_status_reg(13 downto 12);
				error_status_read(ADC_CLK_SOURCES+7 downto 0) <= error_status_reg(ADC_CLK_SOURCES+7 downto 0);
				-- Clear the sticky error status bits by updating to the instantaneous status.
				error_status_reg(13 downto 12) <= error_monitor(13 downto 12);
				error_status_reg(ADC_CLK_SOURCES+7 downto 0) <= error_monitor(ADC_CLK_SOURCES+7 downto 0);
			else
				-- Update the error status using sticky bits.
				error_status_reg(13 downto 12) <= error_status_reg(13 downto 12) or error_monitor(13 downto 12);
				error_status_reg(ADC_CLK_SOURCES+7 downto 0) <= error_status_reg(ADC_CLK_SOURCES+7 downto 0) or error_monitor(ADC_CLK_SOURCES+7 downto 0);
			end if;
			
            -- Write to the serial payload register at address x"0020".
            if bar0_write = '1' and ("00" & slv_addr(13 downto 3) & "000" = x"0020") then
				if bytevalid(1) = '1' then serial_payload_reg(63 downto 32) <= slv_dataout(63 downto 32); end if;
				if bytevalid(0) = '1' then serial_payload_reg(31 downto 0) <= slv_dataout(31 downto 0); end if;
            elsif serial_shift = '1' then
            	-- The serial payload register is also a serial shift register.
				serial_payload_reg <= serial_payload_reg(62 downto 0) & serial_read;
			end if;
                          
        end if;
    end process;
    
    -- Set unused clock status bits to one.
	CLK_SRC: if ADC_CLK_SOURCES < 4 generate
		error_status_reg(11 downto ADC_CLK_SOURCES+8) <= error_monitor(11 downto ADC_CLK_SOURCES+8);
		error_status_read(11 downto ADC_CLK_SOURCES+8) <= error_status_reg(11 downto ADC_CLK_SOURCES+8);
	end generate;

    -- This process needs to be independent of the master reset to hold the software reset active through the stretch count.
    process (clk)
    begin
        if rising_edge(clk) then
        	-- Stretch the software reset to meet the minimum pulse width requirement of the IDELAYCTRL (52 ns).
			if software_reset_reg = '1' or software_reset_stretch /= x"0" then
				software_reset_stretch <= software_reset_stretch + x"1";
			end if;
        end if;
    end process;
   
--***************************************************** READ TRANSACTIONS ******************************************************

    process (clk,rst)
    begin
        if rst = '1' then
        
			bar0_address_error(1) <= '0';
			bar0_rddata <= (others => '0');
			clk_lck_busy_reg <= '0';
			
        elsif rising_edge(clk) then
  
            -- Target read decoder that loads selected register contents onto the DMA bus for a target read completion.
            if dma_rdchannel = '1' and dma_rd = '1' and bar0_read = '1' then
            
            	-- Set the default value to zero for cases that do not assign all of the bits.
				bar0_rddata <= (others => '0');
				
                case "00" & dma_rdaddr & "000" is
                
                	--  x"0000" is a write only register.
                	
                	-- Clock initialization status.
					when x"0008" =>
						bar0_rddata(AXI_WIDTH-64) <= clk_lck_busy_reg;
						-- Clear the sticky bit.
						clk_lck_busy_reg <= '0';
											
					-- Error status.
                    when x"0010" =>
						bar0_rddata(13 downto 0) <= error_status_read;
                 
					-- Error interrupt mask register settings.
                    when x"0018" =>
						bar0_rddata((AXI_WIDTH-64)+13 downto (AXI_WIDTH-64)) <= error_int_mask_reg;
                 
					-- Serial bus payload register.
                    when x"0020" =>
						bar0_rddata(63 downto 0) <= serial_payload_reg;

					-- Serial bus configuration register.
                    when x"0030" =>
						bar0_rddata(9 downto 0) <= serial_config_reg;

					-- Serial bus status.
                    when x"0038" =>
                        bar0_rddata(AXI_WIDTH-64) <= serial_busy;

					-- BPI reply pending.
                    when x"0040" =>
						bar0_rddata(0) <= bpi_cmd_full;

 					-- BPI reply bus.
                    when x"0048" =>
						bar0_rddata((AXI_WIDTH-64)+32 downto (AXI_WIDTH-64)) <= bpi_reply_bus(32 downto 0);

					-- GPIO configuration register.
                    when x"0050" =>
						bar0_rddata(GPIO_COUNT downto 0) <= gpio_config_reg;

					-- USER configuration register.
                    when x"0058" =>
                        bar0_rddata((AXI_WIDTH-64)+USER_COUNT downto (AXI_WIDTH-64)) <= user_config_reg;

					-- Global status.
					when x"0100" =>
						bar0_rddata(32) <= global_status(0);
						bar0_rddata(12) <= global_status(4);
						bar0_rddata(8) <= global_status(3);
						bar0_rddata(4) <= global_status(2);
						bar0_rddata(0) <= global_status(1);
						
					-- Global interrupt mask register.
					when x"0108" =>
						bar0_rddata(AXI_WIDTH-64) <= global_int_mask_reg;
						
					-- Measured interrupt response time.
					when x"0110" =>
						bar0_rddata(27 downto 0) <= int_response;
						
					-- Interrupt timeout value.
					when x"0118" =>
						bar0_rddata((AXI_WIDTH-64)+27 downto (AXI_WIDTH-64)) <= int_timeout_reg;
						
					-- DMA channel 1 and channel 2 status.
					when x"1F00" =>
						bar0_rddata(63 downto 48) <= dma_current_burst_local(2);
						bar0_rddata(46 downto 44) <= dma_engine(2);
						bar0_rddata(41 downto 32) <= dma_current_page_local(2);
						bar0_rddata(31 downto 16) <= dma_current_burst_local(1);
						bar0_rddata(14 downto 12) <= dma_engine(1);
						bar0_rddata(9 downto 0) <= dma_current_page_local(1);
						
					-- DMA channel 3 and channel 4 status.
					when x"1F08" =>
						bar0_rddata((AXI_WIDTH-64)+63 downto (AXI_WIDTH-64)+48) <= dma_current_burst_local(4);
						bar0_rddata((AXI_WIDTH-64)+46 downto (AXI_WIDTH-64)+44) <= dma_engine(4);
						bar0_rddata((AXI_WIDTH-64)+41 downto(AXI_WIDTH-64)+32) <= dma_current_page_local(4);
						bar0_rddata((AXI_WIDTH-64)+31 downto (AXI_WIDTH-64)+16) <= dma_current_burst_local(3);
						bar0_rddata((AXI_WIDTH-64)+14 downto (AXI_WIDTH-64)+12) <= dma_engine(3);
						bar0_rddata((AXI_WIDTH-64)+9 downto (AXI_WIDTH-64)) <= dma_current_page_local(3);
						
					-- DMA channel 5 and channel 6 status.
					when x"1F10" =>
						bar0_rddata(63 downto 48) <= dma_current_burst_local(6);
						bar0_rddata(46 downto 44) <= dma_engine(6);
						bar0_rddata(41 downto 32) <= dma_current_page_local(6);
						bar0_rddata(31 downto 16) <= dma_current_burst_local(5);
						bar0_rddata(14 downto 12) <= dma_engine(5);
						bar0_rddata(9 downto 0) <= dma_current_page_local(5);
						
					-- DMA channel 7 and channel 8 status.
					when x"1F18" =>
						bar0_rddata((AXI_WIDTH-64)+63 downto (AXI_WIDTH-64)+48) <= dma_current_burst_local(8);
						bar0_rddata((AXI_WIDTH-64)+46 downto (AXI_WIDTH-64)+44) <= dma_engine(8);
						bar0_rddata((AXI_WIDTH-64)+41 downto(AXI_WIDTH-64)+32) <= dma_current_page_local(8);
						bar0_rddata((AXI_WIDTH-64)+31 downto (AXI_WIDTH-64)+16) <= dma_current_burst_local(7);
						bar0_rddata((AXI_WIDTH-64)+14 downto (AXI_WIDTH-64)+12) <= dma_engine(7);
						bar0_rddata((AXI_WIDTH-64)+9 downto (AXI_WIDTH-64)) <= dma_current_page_local(7);
						
					-- Index into the DMA page address and descriptor memories.
					when x"1FF8" =>
						bar0_rddata((AXI_WIDTH-64)+15 downto (AXI_WIDTH-64)) <= dma_bram_select_reg;
						
					-- Return an error if the address is not valid.
                    when others  =>
						bar0_rddata(31 downto 0) <= x"badaddb0";
						if dma_rdaddr(13) = '0' then bar0_address_error(1) <= '1'; end if;
						
                end case;
			else
				-- Create a sticky bit to hold the clock lock active status until it can be read.
				if clk_lck_busy = '1' then
					clk_lck_busy_reg <= '1';
				end if;
				bar0_address_error(1) <= '0';
            end if;
            
        end if;
    end process;
       
--***************************************************** SIGNAL ASSIGNMENTS *****************************************************

	software_reset <= '1' when software_reset_stretch /= x"0" else '0';
	clk_lck_start <= '1' when clk_lck_stretch /= x"0" else '0';
    error_status <= error_status_reg;
    error_int_mask <= error_int_mask_reg;
    global_int_mask <= global_int_mask_reg;
	int_timeout <= int_timeout_reg;
    serial_config <= serial_config_reg;
    serial_payload <= serial_payload_reg(63) & serial_payload_reg(55) & serial_payload_reg(47) & serial_payload_reg(39) & serial_payload_reg(31) & serial_payload_reg(23) & serial_payload_reg(15) & serial_payload_reg(7);
    gpio_direction <= gpio_config_reg(GPIO_COUNT downto 1);
    gpio_tristate <= gpio_config_reg(0);
    user_direction <= user_config_reg(USER_COUNT downto 1);
    user_tristate <= user_config_reg(0);
	bpi_cmd_bus <= bpi_cmd_bus_reg;

--************************************ DMA PAGE ADDRESS AND DESCRIPTOR MEMORY TRANSACTIONS *************************************

	-- Decode target write transactions to the DMA page address and descriptor memories.
	dma_bram_we <= bar0_write and slv_addr(13);
	
	-- Convert the byte valid flag bits to standard logic vector so that they match the BRAM port type.
	dma_bram_bytevalid(0)(0) <= bytevalid(0);
	dma_bram_bytevalid(1)(0) <= bytevalid(1);
	
	-- Lane steering depends on the width of the AXI bus.
	dma_bram_data(1) <= slv_dataout(63 downto 32) when slv_addr(3) = '0' else slv_dataout((AXI_WIDTH-64)+63 downto (AXI_WIDTH-64)+32);
	dma_bram_data(0) <= slv_dataout(31 downto 0) when slv_addr(3) = '0' else slv_dataout((AXI_WIDTH-64)+31 downto (AXI_WIDTH-64));
	
	-- Generate a 64-bit page address memory for each DMA channel.
	PAGE_MEMORY : for i in 1 to DMA_CHANNEL_COUNT generate
		dma_bram_page_we(i) <= dma_bram_we and dma_bram_select_reg(i);
			-- Use two 32-bit BRAMs to emulate a register write operation.
			BRAM_X2 : for j in 0 to 1 generate
				PAGE_BRAM : entity work.dma_on_demand_memory
					port map(
						clka => clk,
						ena => dma_bram_page_we(i),
						wea => dma_bram_bytevalid(j),
						addra => slv_addr(12 downto 3),
						dina => dma_bram_data(j),
						clkb => clk,
						addrb => dma_current_page(i),
						doutb => dma_page_data(i)((32*j)+31 downto 32*j)
						);
			end generate;
		dma_page_address(i) <= dma_page_data(i)(63 downto 32) & dma_page_data(i)(31 downto 0);
	end generate;
			
	-- Generate a 32-bit descriptor memory for each DMA channel.
	DESC_MEMORY : for i in 1 to DMA_CHANNEL_COUNT generate
		dma_bram_desc_we(i) <= dma_bram_we and dma_bram_select_reg(i+8);
		DESC_BRAM : entity work.dma_on_demand_memory
			port map(
				clka => clk,
				ena => dma_bram_desc_we(i),
				wea => dma_bram_bytevalid(0),
				addra => slv_addr(12 downto 3),
				dina => dma_bram_data(0),
				clkb => clk,
				addrb => dma_current_page(i),
				doutb => dma_desc_data(i)
				);
		dma_descriptor(i) <= dma_desc_data(i);
	end generate;

	-- Make signal assignments to unused DMA channels.
	DMA_CH : for i in 1 to 8 generate
		CH_ACTIVE: if i <= DMA_CHANNEL_COUNT generate
			dma_current_page_local(i) <= dma_current_page(i);
			dma_current_burst_local(i) <= dma_current_burst(i);
		end generate;	
		CH_UNUSED: if i > DMA_CHANNEL_COUNT generate
			dma_current_page_local(i) <= (others=>'0');
			dma_current_burst_local(i) <= (others=>'0');
		end generate;	
	end generate;	
		
end structural;
