--******************************************************************************************************************************
--
-- Description: BPI configuration flash programmer that can be commanded from software.
--					
-- Rev History: 05-12-2015:	Added revision select command and substituted IP generated MMCM with macro equivalent.
--				04-17-2015:	Added functionality to program flash through the PCI interface.
--				01-29-2014:	Initial release.
--
--******************************************************************************************************************************

library ieee;
use ieee.std_logic_1164.all;
use IEEE.numeric_std.all;

Library UNISIM;
use UNISIM.vcomponents.all;
library UNIMACRO;
use unimacro.Vcomponents.all;

use work.model_specific_package.all;

entity bpi_flash is
	port (
    -- Reset and clock.                                                                                      
	rst 			: in std_logic;								-- Active high asynchronous reset.
	user_clk 		: in std_logic;								-- System clock.
	clk200 			: in std_logic;								-- 200 MHz clock from an external oscillator.

	-- Connections to command/status registers in the BAR0 module.
	bpi_cmd_bus		: in std_logic_vector(32 downto 0);			-- Command bus that carries instructions and data to program the BPI flash.
	bpi_cmd_full	: out std_logic; 							-- Logic one indicates the BPI command FIFO cannot accept another burst.
	bpi_reply_bus	: out std_logic_vector(32 downto 0); 		-- Reply bus that returns data and errors from the BPI flash.
	bpi_reply_rd	: in std_logic; 							-- Logic one pops the BPI reply FIFO to the next value.
	clk20_locked	: out std_logic; 							-- Logic one indicates MMCM locked.
	
	--	The following signals connect to external devices through FPGA pins.
	bpi_a			: out std_logic_vector(BPI_ADD-1 downto 0);	-- Address bus.
	bpi_rs			: out std_logic_vector(1 downto 0);			-- Revision address.
	bpi_rs_disable	: out std_logic; 							-- Revision select disable.
	bpi_d_in		: in std_logic_vector(15 downto 0);			-- Data input (read) bus.
	bpi_d_out		: out std_logic_vector(15 downto 0);		-- Data output (write) bus.
	bpi_d_tristate	: out std_logic;							-- Active high data bus tristate.
	bpi_adv_b		: out std_logic; 							-- Active low address valid.
	bpi_fcs_b		: out std_logic; 							-- Active low chip enable.
	bpi_fwe_b		: out std_logic; 							-- Active low write enable.
	bpi_foe_b		: out std_logic; 							-- Active low output enable.
	bpi_vpp			: out std_logic 							-- Active high enable voltage for program and erase operations.
	);
end bpi_flash;

architecture structural of bpi_flash is

	type PROGRAM_STATE is (idle, enable, revision, address, write, read1, read2, read3, pop);	--	Type definitions for the state machine.
	signal bpi_state 	: PROGRAM_STATE;			-- Current state.
	signal command		: std_logic_vector(31 downto 0);	-- Command bus that carries instructions and data to program the flash.
	signal command_rd_en: std_logic;						-- Command bus read enable.
	signal command_vld	: std_logic;						-- Logic one indicates a valid command on the bus.
	signal reply		: std_logic_vector(31 downto 0);	-- Reply bus that returns data and errors from the BPI flash.
	signal reply_wr_en	: std_logic;						-- Reply bus write enable.
	signal reply_vld	: std_logic;						-- Logic one indicates a valid reply on the bus.
	signal reply_full	: std_logic;						-- Logic zero indicates that the reply FIFO can accept another burst.
	signal clkfbout		: std_logic;						-- MMCM feedback output.
	signal clkfbin		: std_logic;						-- MMCM feedback output on a global clock buffer.
	signal clkout0		: std_logic;						-- MMCM input clock divided by 10.
	signal clk20		: std_logic;						-- 20 MHz clock on a global clock bufffer.
	
begin

--************************************************ INSTANTIATE CLOCK AND FIFO **************************************************

    -- Divide the 200 MHz clock by ten to produce a 20 MHz clock for the programmer.
	CLK200_CLK20 : MMCME2_BASE
	generic map (
		CLKFBOUT_MULT_F => 42.375,
		CLKIN1_PERIOD => 5.0,
		CLKOUT0_DIVIDE_F => 42.375,
		DIVCLK_DIVIDE => 10
	)
	port map (
		CLKFBIN => clkfbin,
		CLKFBOUT => clkfbout,
		CLKFBOUTB => open,
		CLKOUT0 => clkout0,
		CLKOUT0B => open,
		CLKOUT1 => open,
		CLKOUT1B => open,
		CLKOUT2 => open,
		CLKOUT2B => open,
		CLKOUT3 => open,
		CLKOUT3B => open,
		CLKOUT4 => open,
		CLKOUT5 => open,
		CLKOUT6 => open,
		LOCKED => clk20_locked,
		CLKIN1 => clk200,
		PWRDWN => '0',
		RST => rst
	);
	-- Map the MMCM feedback to a global clock buffer.
	CLKFB_BUFG : BUFG port map (I => clkfbout, O => clkfbin);
	-- Map the MMCM divided output to a 20 MHz global clock buffer.
	CLKOUT0_BUFG : BUFG port map (I => clkout0, O => clk20);	
	
    -- FIFO to buffer incoming commands and data.
   	COMMAND_FIFO : entity work.fifo_flash_512x32
	port map (
		rst 		=> rst,
		wr_clk 		=> user_clk,
		rd_clk 		=> clk20,
		din 		=> bpi_cmd_bus(31 downto 0),
		wr_en 		=> bpi_cmd_bus(32),
		rd_en		=> command_rd_en,
		dout		=> command,
		full		=> open,
		empty		=> open,
		valid		=> command_vld,
		prog_full	=> bpi_cmd_full
	);
	
    -- FIFO to buffer outgoing data and error messages.
   	REPLY_FIFO : entity work.fifo_flash_512x32
	port map (
		rst 		=> rst,
		wr_clk 		=> clk20,
		rd_clk 		=> user_clk,
		din 		=> reply,
		wr_en 		=> reply_wr_en,
		rd_en		=> bpi_reply_rd,
		dout		=> bpi_reply_bus(31 downto 0),
		full		=> open,
		empty		=> open,
		valid		=> bpi_reply_bus(32),
		prog_full	=> reply_full
	);
	
	--************************************************* STATE MACHINE TRANSITIONS **************************************************

	process (clk20,rst)
    begin
		if rst='1' then
		
			bpi_state <= idle;

        elsif rising_edge(clk20) then
        
			case bpi_state is
				-- Remain idle until a command is available.
				when idle =>
					-- Monitor the command/data FIFO for a new command.
					if command_vld = '1' then
						case command(31 downto 28) is
							when x"0" => bpi_state <= enable;
							when x"1" => bpi_state <= revision;
							when x"2" => bpi_state <= address;
							when x"3" => bpi_state <= write;
							when x"4" => bpi_state <= read1;
							when others => null;
						end case;
					end if;
					
				-- Turn the program voltage and chip select on (1) or off (0).
				-- Note: This control is a hardware build option that is only present if requested.
				when enable =>
					bpi_state <= pop;
					
				-- Store the revision select settings.
				when revision =>
					bpi_state <= pop;
					
				-- Store the address settings.
				when address =>
					bpi_state <= pop;
					
				-- Perform a command write operation.
				when write =>
					bpi_state <= pop;
					
				-- Tri-state the FPGA data outputs.
				when read1 =>
					bpi_state <= read2;
				-- Enable the BPI outputs while FPGA data outputs are tri-state.
				when read2 =>
					-- Do not continue if the reply FIFO is full.
					if reply_full = '0' then
						bpi_state <= read3;
					else
						bpi_state <= read2;
					end if;
				-- Store the read data and disable the BPI outputs while FPGA data outputs are tri-state.
				-- Note: Bus will turn-around back to write flash during the next state (pop).
				when read3 =>
					bpi_state <= pop;
					
				-- Pop the command FIFO to receive the next instruction or data.
				when pop =>
					bpi_state <= idle;
					
				when others  =>
					bpi_state <= idle;
			end case;
        end if;
    end process;

--*************************************************** STATE MACHINE EVENTS *****************************************************

	process (clk20,rst)
    begin
		if rst='1' then
		
			bpi_vpp <= '0';
			bpi_fcs_b <= '1';
			bpi_a <= (others => '0');
			bpi_rs <= (others => '0');
			bpi_rs_disable <= '1';
			bpi_d_out <= (others => '0');
			bpi_fwe_b <= '1';
			bpi_d_tristate <= '0';
			bpi_foe_b <= '1';
			reply(15 downto 0) <= (others => '0');
			reply_wr_en <= '0';

        elsif rising_edge(clk20) then
        
			-- Turn the program voltage and chip select on (1) or off (0).
			-- Note: This control is a hardware build option that is only present if requested.
			if bpi_state = enable then
				-- Logic one asserts the program voltage and chip select, zero deasserts.
				bpi_vpp <= command(0);
				bpi_fcs_b <= not(command(0));
			end if;
					
			-- Store the BPI revision select settings when commanded.
			if bpi_state = revision then
				-- Assign the revision select bits.
				bpi_rs <= command(1 downto 0);
				bpi_rs_disable <= not(command(2));
			end if;
			
			-- Store the BPI address when commanded.
			if bpi_state = address then
				-- Assign the address bits.
				bpi_a <= command(22 downto 0);
			end if;
			
 			-- Send a write command to the flash.
			if bpi_state = write then
				-- Assign the data bits.
				bpi_d_out <= command(15 downto 0);
				-- Set the write enable active low.
				bpi_fwe_b <= '0';
			else
				bpi_fwe_b <= '1';
			end if;
					
 			-- Tristate the FPGA data bus throughout a flash read transaction.
			if bpi_state = read1 or bpi_state = read2 or bpi_state = read3 then
				bpi_d_tristate <= '1';
			else
				bpi_d_tristate <= '0';
			end if;
					
 			-- Enable the BPI data bus during a flash read transaction.
			if bpi_state = read2 then
				bpi_foe_b <= '0';
			else
				bpi_foe_b <= '1';
			end if;
			
			-- Return the results from a read operation.
			if bpi_state = read3 then
				reply(15 downto 0) <= bpi_d_in;
				reply_wr_en <= '1';
			else
				reply_wr_en <= '0';
			end if;
			
       end if;
    end process;

    -- Tie the reply MSBs to zero.
    reply(31 downto 16) <= (others => '0');
    
 	-- Pop the current instruction or data off the command FIFO.
    command_rd_en <= '1' when bpi_state = pop else '0';
    
    -- Address values are registered, so tie the control active.
    bpi_adv_b <= '0';

end structural;
