--*******************************************************************************************************************************
--
-- Description: This module is used to assign source clocks to global buffers and connect them to desired inputs.
--
-- Rev History: 08-04-2016	Added MMCM to generate spread spectrum QDR II+ SRAM clock and removed unused signals.
-- 				06-22-2016	Change type of rx_clk_in and tx_clk_in ports.
-- 				08-20-2014	Removed clk500 from I/O ports.
--				09-04-2013	Initial release.
--
--*******************************************************************************************************************************

library ieee;
use ieee.std_logic_1164.all;
use IEEE.numeric_std.all;

Library UNISIM; 
use UNISIM.Vcomponents.all; 

use work.model_specific_package.all;
use work.transceiver_package.all;

entity clock_route is 
	port (
	-- Reset and clocks.
	rst         	: in std_logic;		-- Active high asynchronous reset.       
	sample_clk_in	: in std_logic;		-- IOB sample clock.
    clk200_in		: in std_logic;		-- 200 MHz clock from an external oscillator.
    qdriip_sys_clk	: in std_logic;		-- QDR II+ SRAM system clock.
	
	-- Output reset and clocks.
	rx_clk			: out std_logic;	-- Clock assigned to the receiver datapath.
	tx_clk			: out std_logic;	-- Clock assigned to the transmitter datapath.
    clk200			: out std_logic;		-- 200 MHz clock on a global clock buffer.
    qdriip_clk		: out std_logic;	-- Clock assigned to SRAM-A and SRAM-B.
    qdriip_rst		: out std_logic	-- Active high asynchronous SRAM reset.
	);	
	end clock_route;	

architecture behavioral of clock_route is

	signal sample_clk	: std_logic;	-- Sample clock on global clock buffer.
	signal clkfb_463ss		: std_logic;						-- 463 MHz spread spectrum MMCM feedback.
	signal clkout0_463ss	: std_logic;						-- 463 MHz spread spectrum MMCM output.
	signal lock_463ss		: std_logic;						-- 463 MHz MMCM lock indicator.

begin                    

--**************************************************** GLOBAL CLOCK BUFFERS ****************************************************

	-- Assign the 200 MHz oscillator source to a global clock buffer.
	CLK200_BUFG : BUFG port map (I => clk200_in, O => clk200);
	
	-- Assign the sample clock source to a global clock buffer.
	SAMPLE_CLK_BUFG : BUFG port map (I => sample_clk_in, O => sample_clk);
		
--************************************************* RECEIVER CLOCK ASSIGNMENTS *************************************************

	-- Connect the receiver clock to sample clock if requested.
	RX_CLK_SAMPLE: if RX_CHANNEL_COUNT > 0 and RX_CLK_SELECT = 0 generate
		rx_clk <= sample_clk;
	end generate;
	
	-- Tie the receiver clock low if there are no receiver channels.
	RX_CLK_NONE: if RX_CHANNEL_COUNT = 0 generate
		rx_clk <= '0';
	end generate;
	
--*********************************************** TRANSMITTER CLOCK ASSIGNMENTS ************************************************

	-- Connect the transmitter clock to sample clock if requested.
	TX_CLK_SAMPLE: if TX_CHANNEL_COUNT > 0 and TX_CLK_SELECT = 0 generate
		tx_clk <= sample_clk;
	end generate;
	
	-- Tie the transmitter clock low if there are no transmitter channels.
	TX_CLK_NONE: if TX_CHANNEL_COUNT = 0 generate
		tx_clk <= '0';
	end generate;
	
--************************************************* QDR SRAM CLOCK ASSIGNMENTS *************************************************

	-- Route the 450 MHz QDR II+ system clock directly to the SRAM.
	QDR_CLK_450: if QDR_ADDR_WIDTH > 0 and QDR_SPREAD = false generate
		qdriip_clk <= qdriip_sys_clk;
		qdriip_rst <= rst;
	end generate;
	
	-- Create a spread spectrum 463 MHz SRAM clock from the 150 MHz QDR II+ system clock.
	QDR_CLK_463: if QDR_ADDR_WIDTH > 0 and QDR_SPREAD = true generate
		-- Instantiate a spread spectrum clock (488.89 MHz center).
		CLK463 : MMCME2_ADV
		generic map (
			BANDWIDTH => "LOW",
			CLKFBOUT_MULT_F => 22.000,
			CLKIN1_PERIOD => 6.666,
--			CLKOUT0_DIVIDE_F => 2.25,
			CLKOUT0_DIVIDE_F => 2.375,
			DIVCLK_DIVIDE => 3,
			SS_EN => "TRUE",
			SS_MOD_PERIOD => 4000,
			SS_MODE => "CENTER_HIGH"
		)
		port map (
			CLKOUT0 => clkout0_463ss,
			CLKOUT0B => open,
			CLKOUT1 => open,
			CLKOUT1B => open,
			CLKOUT2 => open,
			CLKOUT2B => open,
			CLKOUT3 => open,
			CLKOUT3B => open,
			CLKOUT4 => open,
			CLKOUT5 => open,
			CLKOUT6 => open,
			DO => open,
			DRDY => open,
			PSDONE => open,
			CLKFBOUT => clkfb_463ss,
			CLKFBOUTB => open,
			CLKFBSTOPPED => open,
			CLKINSTOPPED => open,
			LOCKED => lock_463ss,
			CLKIN1 => qdriip_sys_clk,
			CLKIN2 => '0',
			CLKINSEL => '1',
			PWRDWN => '0',
			RST => rst,
			DADDR => "0000000",
			DCLK => '0',
			DEN => '0',
			DI => x"0000",
			DWE => '0',
			PSCLK => '0',
			PSEN => '0',
			PSINCDEC => '0',
			CLKFBIN => clkfb_463ss
		);
		-- Map the MMCM divided output to a global clock buffer.
		CLKOUT0_463SSBUFG : BUFG port map (I => clkout0_463ss, O => qdriip_clk);
		-- Use the lock indicator to reset the spread spectrum clock generator.
		qdriip_rst <= not(lock_463ss);
	end generate;

	-- Assign unused signals if there is no QDR II+ SRAM.
	NO_SRAM : if QDR_ADDR_WIDTH = 0 generate	
		qdriip_clk <= '0';
		qdriip_rst <= '0';
	end generate;

end behavioral;
