--*******************************************************************************************************************************
--
-- Description: This core is used to perform read and write operations to a single QDR II+ SRAM. The details of
--				the SRAM hardware interface are managed by the core.  The application logic connects to a simplified
--				dual-port memory structure with separate read and write controls.  Each port has a dedicated clock,
--				address bus, data bus, byte enable, and data valid enable.
--
--				The clock_lock core is used to optimize the FPGA IO timing for read operations from the SRAM.
--				A clock forwarding scheme is used to ensure that timing contraints are met during write operations
--				to the SRAM.
--				
--				A write operation is peformed by simultaneously presenting the address (wr_add), data (wr_data), and
--				write enable (wr_en) signals to the core.  The byte enable input (wr_ben) controls which bytes, if any,
--				are masked to the SRAM.
--
--				A read operation is initiated by supplying the desired address (rd_add) while driving the read enable
--				(rden) signal active.  The data ready (rd_vld) signal will indicate when the contents of the requested
--				address become available on the data (rd_data) port.  The output data is held until the cycle is completed
--				with an acknowledgement (rd_ack), allowing the results of the next read operation to show through.
--
-- Rev History: 08-04-2016:	Changed the name of qdriip_sys_clk to qdriip_clk to distinguish spread spectrum clock name.
-- 				07-14-2015:	Completely new code to replace the MIG generated QDR II+ SRAM interface.
-- 				05-21-2014:	Consolidated QDR II+ SRAM registers and added control signal for the core calibration function.
--				10-07-2013: Initial Release.
--
--*******************************************************************************************************************************

library ieee;
use ieee.std_logic_1164.all;
use IEEE.numeric_std.all; 

Library UNISIM; 
use UNISIM.Vcomponents.all; 

use work.model_specific_package.all;
use work.transceiver_package.all;

entity qdr_sram is
	port (
	-- Reset, clocks and enable supplied by the application logic.
	rst         		: in std_logic;		-- Active high reset, logic one places all registers in a reset state.	
	qdriip_clk			: in std_logic;		-- QDR II+ fixed frequency or spread spectrum clock.
	clk200 				: in std_logic;		-- 200 MHz clock on global clock buffer.
	dll_on      		: in std_logic;		-- Logic one enables the delay-locked loop (DLL) internal to the SRAM.
	
	-- Write transaction interface signals.
	wr_clk       		: in std_logic;		-- Clock assigned to the write port.
	wr_rdy       		: out std_logic;	-- Logic one indicates that the write port is ready to accept another command.
	wr_cmd        		: in std_logic;		-- Logic one commits a write cycle to the SRAM using the current address and data.
	wr_add      		: in std_logic_vector(QDR_ADDR_WIDTH-1 downto 0);	-- Address input to the write port. 
	wr_data     		: in std_logic_vector(71 downto 0);	-- Data input to the write port.
	wr_bw	    		: in std_logic_vector(7 downto 0);	-- Logic one enables individual bytes [B7 ... B0] to be written.
	
	-- Read transaction interface signals.
	rd_clk       		: in std_logic;		-- Clock assigned to the read port.
	rd_rdy        		: out std_logic;	-- Logic one indicates that the read port is ready to accept another command.
	rd_cmd        		: in std_logic;		-- Logic one commits a read cycle to the SRAM using the current address.
	rd_add      		: in std_logic_vector(QDR_ADDR_WIDTH-1 downto 0);	-- Address input to the read port.
	rd_data      		: out std_logic_vector(71 downto 0);				-- Data output from the read port.
	rd_vld	    		: out std_logic;	-- Logic one indicates that data is available from the requested read operation.
	rd_ack       		: in std_logic;	 	-- Logic one releases the current data and allows the next requested value to show through.
											-- The data ready (rd_vld) flag will remain high if new data is available on the next clock.
									
	-- Clock lock command and status signals.
	dll_rst_start		: in std_logic; 	-- Connect to the ready output of the IDELAY CONTROLLER.
	clk_lck_status		: out std_logic; 	-- Logic one indicates that clocks are locked.
	
	-- The following signals connect to a QDR II+ SRAM device through FPGA pins.
	qdriip_cq			: in  std_logic;						-- QDR II+ SRAM echo clock.
	qdriip_q			: in  std_logic_vector(17 downto 0);	-- QDR II+ SRAM synchronous read data outputs.
	qdriip_qvld			: in  std_logic;						-- QDR II+ SRAM read data valid flag.
	qdriip_k			: out std_logic;						-- QDR II+ SRAM input clock.
	qdriip_d			: out std_logic_vector(17 downto 0);	-- QDR II+ SRAM synchronous write data inputs.
	qdriip_sa			: out std_logic_vector(QDR_ADDR_WIDTH-1 downto 0);	-- QDR II+ SRAM synchronous address inputs.
	qdriip_w_n			: out std_logic;						-- QDR II+ SRAM synchronous active low write enable.
	qdriip_r_n			: out std_logic;						-- QDR II+ SRAM synchronous active low read enable.
	qdriip_bw_n			: out std_logic_vector(1 downto 0);		-- QDR II+ SRAM synchronous byte enables.
	qdriip_dll_off_n	: out std_logic							-- QDR II+ SRAM active low DLL disable.
	);	
end qdr_sram;

architecture behavioral of qdr_sram is

    -- State machine signals.
	type QDRII_STATE is (idle, read, write);	-- Type definitions for the QDR II+ state machine.
	signal qdr_state 		: QDRII_STATE;		-- Current state.
	
	type TRANS_STATE is (idle, beat12, beat34);	-- Type definitions for the transaction state machine.
	signal wr_state 		: TRANS_STATE;		-- Current write state.
	signal rd_state 		: TRANS_STATE;		-- Current read state.
	
    -- Write port signals.
	signal wr_burst_data	: std_logic_vector(71 downto 0);	-- Burst of four (b4) write data.
	signal wr_descriptor	: DUAL_36;							-- Descriptor consisting of address and byte enables for write transactions.
	signal wr_desc_pop		: std_logic;						-- Control bit to pop the write address FIFO to the next value.
	signal wr_data_pop		: std_logic_vector(1 downto 0);		-- Control bit to pop the write data FIFO to the next value.
	signal wr_desc_vld		: std_logic;						-- Valid flags from the write descriptor (0) and data (1) FIFOs.
	signal wr_full			: std_logic_vector(1 downto 0);		-- Full flags from the write data FIFOs.
	signal wr_valid			: std_logic_vector(1 downto 0);		-- Logic one indicates that data is available from the write data FIFO.
	signal ddr_k			: std_logic;						-- Forwarded write clock (K) in DDR format.
	signal sa				: std_logic_vector(QDR_ADDR_WIDTH-1 downto 0);	-- Either a read or write address (SA) in SDR format.
	signal ddr_sa			: std_logic_vector(QDR_ADDR_WIDTH-1 downto 0);	-- Either a read or write address (SA) in DDR format.
	signal w_n				: std_logic;						-- Active low write enable (WN) in SDR format.
	signal ddr_wn			: std_logic;						-- Active low write enable (WN) in DDR format.
	signal d				: std_logic_vector(35 downto 0);	-- Write data (D) in SDR format.
	signal ddr_d			: std_logic_vector(17 downto 0);	-- Write data (D) in DDR format.
	signal bwn_hold			: std_logic_vector(7 downto 0);		-- Storage for byte enables (BWN) that accompany write addresses.
	signal bw_n				: std_logic_vector(3 downto 0);		-- Byte enables (BWN) in SDR format.
	signal ddr_bwn			: std_logic_vector(1 downto 0);		-- Byte enables (BWN) in DDR format.
	
    -- Read port signals.
	signal rd_descriptor	: DUAL_36;							-- Descriptor consisting of address for read transactions.
	signal rd_desc_pop		: std_logic;						-- Control bit to pop the read address FIFO to the next value.
	signal rd_desc_vld		: std_logic;						-- Valid flag from the read descriptor FIFO.
	signal rd_full			: std_logic;						-- Full flag from the read address FIFO.
	signal r_n				: std_logic;						-- Active low read enable (RN) in SDR format.
	signal ddr_rn			: std_logic;						-- Active low read enable (RN) in DDR format.
	signal ddr_qvld			: std_logic;						-- Data valid flag (QVLD) in DDR format.
	signal qvld				: std_logic_vector(1 downto 0);		-- Data valid flag (QVLD) in SDR format.
	signal ddr_q			: std_logic_vector(17 downto 0);	-- Read data (Q) in DDR format.
	signal q				: std_logic_vector(35 downto 0);	-- Read data (Q) in SDR format.
	signal ddr_cq			: std_logic;						-- Echo clock (CQ) in DDR format.
	signal cq				: std_logic_vector(0 downto 0);		-- Echo clock (CQ) in DR format.
	signal rd_split_data	: std_logic_vector(17 downto 0);	-- Storage for a single data word that is not paired in a DDR cycle.
	signal rd_burst_data	: std_logic_vector(35 downto 0);	-- Burst of four (b4) read data.
	signal rd_data_load		: std_logic_vector(1 downto 0);		-- Logic one stores read data in FIFO.
	signal rd_valid			: std_logic_vector(1 downto 0);		-- Logic one indicates that data is available from the read data FIFO.
	
    -- DLL reset and clock lock signals.
	signal dll_rst_busy		: std_logic;						-- Logic one indicates that the DLL reset process is active.
	signal dll_rst_done		: std_logic;						-- Single period lock done pulse to indicate that the process is complete.
	signal clk_lck_rst		: std_logic;						-- Active high reset to the clock lock function.
	signal clk_edge_detect	: std_logic_vector(0 downto 0);		-- Indicates which edge was detected by clock lock,  rising (0) or falling (1).
	signal lock_status		: std_logic_vector(0 downto 0);		-- Logic one indicates that a clock lock solution was found.
	signal fwd_counter		: std_logic_vector(4 downto 0);		-- Tap count reported by the IDELAY attached to the forwarded clock.
	signal fwd_ld			: std_logic_vector(0 downto 0);		-- Logic one loads the starting tap value in the IDELAY attached to the forwarded clock.
	signal fwd_ce			: std_logic_vector(0 downto 0);		-- Logic one enables the increment/decrement function in the IDELAY attached to the forwarded clock.
	signal fwd_inc			: std_logic_vector(0 downto 0);		-- Logic one increments the tap value and logic zero decrements the tap value in the IDELAY attached to the forwarded clock.
	signal data_ld			: std_logic_vector(0 downto 0);		-- Logic one loads the starting tap value in the IDELAY attached to the data bits.
	signal data_ce			: std_logic_vector(0 downto 0);		-- Logic one enables the increment/decrement function in the IDELAY attached to the data bits.
	signal data_inc			: std_logic_vector(0 downto 0);		-- Logic one increments the tap value and logic zero decrements the tap value in the IDELAY attached to the data bits.

begin

--***************************************** INSTANTIATE WRITE ADDRESS AND DATA FIFOS *******************************************

	-- Build a write descriptor consisting of the address and byte enables.
	wr_descriptor(0)(35 downto 24) <= x"0" & wr_bw;
	wr_descriptor(0)(23 downto QDR_ADDR_WIDTH) <= (others => '0');
	wr_descriptor(0)(QDR_ADDR_WIDTH-1 downto 0) <= wr_add;

	-- FIFO to supply addresses for write transactions.
	WR_ADD_FIFO : entity work.fifo_qdr_512x36
	port map (
		rst 		=> rst,
		wr_clk 		=> wr_clk,
		rd_clk 		=> qdriip_clk,
		din 		=> wr_descriptor(0),
		wr_en 		=> wr_cmd,
		rd_en		=> wr_desc_pop,
		dout		=> wr_descriptor(1),
		full		=> open,
		empty		=> open,
		valid		=> wr_desc_vld
	);
	
    -- Two FIFOs to supply data for write transactions.
	WR_DATA_FIFO_X2 : for i in 0 to 1 generate
		WR_DATA_FIFO : entity work.fifo_qdr_512x36
		port map (
			rst 		=> rst,
			wr_clk 		=> wr_clk,
			rd_clk 		=> qdriip_clk,
			din 		=> wr_data((36*i)+35 downto 36*i),
			wr_en 		=> wr_cmd,
			rd_en		=> wr_data_pop(i),
			dout		=> wr_burst_data((36*i)+35 downto 36*i),
			full		=> wr_full(i),
			empty		=> open,
			valid		=> wr_valid(i)
		);
	end generate;
	
	-- Generate a write ready flag based on the full status of the address and data FIFOs.	
	wr_rdy <= not(wr_full(0)) and not(wr_full(1));
	
--****************************************** INSTANTIATE READ ADDRESS AND DATA FIFOS *******************************************

	-- Build a read descriptor consisting of the address.
	rd_descriptor(0)(35 downto QDR_ADDR_WIDTH) <= (others => '0');
	rd_descriptor(0)(QDR_ADDR_WIDTH-1 downto 0) <= rd_add;
	
    -- FIFO to supply addresses for read transactions.
	RD_ADD_FIFO : entity work.fifo_qdr_512x36
	port map (
		rst 		=> rst,
		wr_clk 		=> rd_clk,
		rd_clk 		=> qdriip_clk,
		din 		=> rd_descriptor(0),
		wr_en 		=> rd_cmd,
		rd_en		=> rd_desc_pop,
		dout		=> rd_descriptor(1),
		full		=> rd_full,
		empty		=> open,
		valid		=> rd_desc_vld
	);

    -- FIFO to accept data from read transactions.
	RD_DATA_FIFO_X2 : for i in 0 to 1 generate
		RD_DATA_FIFO : entity work.fifo_qdr_512x36
		port map (
			rst 		=> rst,
			wr_clk 		=> qdriip_clk,
			rd_clk 		=> rd_clk,
			din 		=> rd_burst_data,
			wr_en 		=> rd_data_load(i),
			rd_en		=> rd_ack,
			dout		=> rd_data((36*i)+35 downto 36*i),
			full		=> open,
			empty		=> open,
			valid		=> rd_valid(i)
		);
	end generate;
	
	-- Generate a read ready flag based on the full status of the address and data FIFOs.	
	rd_rdy <= not(rd_full);
	-- Generate a read valid flag when data is available at the output of both data FIFOs.	
	rd_vld <= rd_valid(0) and rd_valid(1);
	
--*********************************************** QDR STATE MACHINE TRANSITIONS ************************************************

	process (qdriip_clk,rst)
    begin
		if rst='1' then
		
			qdr_state <= idle;

        elsif rising_edge(qdriip_clk) then
        
			case qdr_state is
				-- Remain idle until a read or write address is available.
				when idle =>
					-- Wait for a valid descriptor before starting a write transaction.
					if  wr_desc_vld = '1' then
						qdr_state <= write;
					-- Wait for a valid descriptor before starting a read transaction.
					elsif rd_desc_vld = '1' then
						qdr_state <= read;
					else
						qdr_state <= idle;
					end if;
					
				-- Perform a write transaction.
				when write =>
					-- Do not advance to the next state until there is data available to start the write transaction.
					if  wr_valid(0) = '1' then
						-- Start a read transaction if there is one pending.
						if rd_desc_vld = '1' then
							qdr_state <= read;
						-- Go back to idle to insert a NOOP between pipelined write transactions.
						else
							qdr_state <= idle;
						end if;
					else
						qdr_state <= write;
					end if;
					
				-- Perform a read transaction.
				when read =>
					-- Start a write transaction if there is one pending.
					if  wr_desc_vld = '1' then
						qdr_state <= write;
					-- Go back to idle to insert a NOOP between pipelined read transactions.
					else
						qdr_state <= idle;
					end if;
					
				when others  =>
					qdr_state <= idle;
			end case;
        end if;
    end process;

--**************************************** WRITE TRANSACTION STATE MACHINE TRANSITIONS *****************************************

	process (qdriip_clk,rst)
    begin
		if rst='1' then
		
			wr_state <= idle;

        elsif rising_edge(qdriip_clk) then
        
			case wr_state is
			
				-- Remain idle until a write transaction begins.
				when idle =>
					-- Wait for a write transaction to begin, but do not advance until data is available.
					if qdr_state = write and wr_valid(0) = '1' then
						wr_state <= beat12;
					end if;
					
				-- Perform the beat 1/2 (rising/falling) data transfers.
				when beat12 =>
					-- Immediately transition to beat 3/4.
					wr_state <= beat34;
					
				-- Perform the beat 3/4 (rising/falling) data transfers.
				when beat34 =>
					-- Return to the beat 1/2 data transfer if a pipelined transaction has already started.
					if  qdr_state = write then
						wr_state <= beat12;
					-- Go back to idle.
					else
						wr_state <= idle;
					end if;
					
				when others  =>
					wr_state <= idle;
			end case;
        end if;
    end process;

--**************************************** READ TRANSACTION STATE MACHINE TRANSITIONS ******************************************

	process (qdriip_clk,rst)
    begin
		if rst='1' then
		
			rd_state <= idle;

        elsif rising_edge(qdriip_clk) then
        
			case rd_state is
			
				-- Remain idle until a read transaction begins.
				when idle =>
					-- Wait for a read transaction to begin.
					if qvld /= "00" then
						rd_state <= beat12;
					end if;
					
				-- Perform the beat 1/2 (rising/falling) data transfers.
				when beat12 =>
					-- Immediately transition to beat 3/4.
					rd_state <= beat34;
					
				-- Perform the beat 3/4 (rising/falling) data transfers.
				when beat34 =>
					-- Return to the beat 1/2 data transfer if a pipelined transaction has already started.
					if  qvld = "11" then
						rd_state <= beat12;
					-- Go back to idle.
					else
						rd_state <= idle;
					end if;
					
				when others  =>
					rd_state <= idle;
			end case;
        end if;
    end process;

--************************************************* WRITE TRANSACTION EVENTS ***************************************************

	process (qdriip_clk,rst)
    begin
		if rst='1' then
		
			sa <= (others => '0');
			bwn_hold <= (others => '0');
			w_n <= '1';
			bw_n <= (others => '0');
			d <= (others => '0');
			wr_data_pop <= (others => '0');

        elsif rising_edge(qdriip_clk) then
        
        	if qdr_state = write then
        		-- Route the write address to the SRAM depending on the transaction type. 
        		sa <= wr_descriptor(1)(QDR_ADDR_WIDTH-1 downto 0);
        		-- Capture the byte enables so that they are still available during the start of another pipelined write transaction.
				bwn_hold <= not(wr_descriptor(1)(31 downto 24));
         		-- Start a write transaction.
         		w_n <= '0';
        	else
        		-- Route the read address to the SRAM depending on the transaction type. 
        		sa <= rd_descriptor(1)(QDR_ADDR_WIDTH-1 downto 0);
        		w_n <= '1';
         	end if;
        
        	if wr_state = beat12 then
				-- Select the lower half of the byte enable FIFO. 
				bw_n <= bwn_hold(3 downto 0);
				-- Select the lower data word FIFO. 
				d <= wr_burst_data(35 downto 0);
				-- Advance the lower write data FIFOs to the next value. 
				wr_data_pop(0) <= '1';
        	else
				-- Select the upper half of the byte enable FIFO. 
				bw_n <= bwn_hold(7 downto 4);
				-- Select the upper data word FIFO. 
				d <= wr_burst_data(71 downto 36);
				wr_data_pop(0) <= '0';
         	end if;
        
        	if wr_state = beat34 then
				-- Advance the upper write data FIFOs to the next value. 
				wr_data_pop(1) <= '1';
        	else
				wr_data_pop(1) <= '0';
         	end if;
                
       end if;
    end process;

    -- Advance the write descriptor FIFO to the next value. 
    wr_desc_pop <= '1' when qdr_state = write else '0';

--************************************************** READ TRANSACTION EVENTS ***************************************************

	process (qdriip_clk,rst)
    begin
		if rst='1' then
		
        	r_n <= '1';
			rd_split_data <= (others => '0');
			rd_burst_data <= (others => '0');
			rd_data_load <= (others => '0');

        elsif rising_edge(qdriip_clk) then
        
        	if qdr_state = read then
         		-- Start a write transaction.
         		r_n <= '0';
        	else
        		-- Route the read address to the SRAM depending on the transaction type. 
        		r_n <= '1';
         	end if;
        
			-- Capture the upper read data word so that it is available on the next clock when dealing with split DDR pairs.
			rd_split_data <= q(35 downto 18);
			
			-- Select which words should be paired for the read data output.
			-- Note: The 2.5 clock cycle latency through the SRAM places the first output on the falling edge of CQ.
			if clk_edge_detect(0) = '1' then
				-- DDR pairs are not split.
				rd_burst_data <= q;
			else
				-- DDR pairs are split.
				rd_burst_data <= q(17 downto 0) & rd_split_data;
			end if;
			
        	if rd_state = beat12 then
        		-- Load the lower read data FIFOs with the next value. 
        		rd_data_load(0) <= '1';
        	else
        		rd_data_load(0) <= '0';
         	end if;
        
        	if rd_state = beat34 then
        		-- Load the upper read data FIFOs with the next value. 
				rd_data_load(1) <= '1';
        	else
				rd_data_load(1) <= '0';
         	end if;
        
       end if;
    end process;
    
    -- Advance the read descriptor FIFO to the next value. 
    rd_desc_pop <= '1' when qdr_state = read else '0';

        
--******************************************** DLL RESET AND CLOCK LOCK CONTROLLER *********************************************

    -- Perform a DLL reset that meets the SRAM timing requirements.
    DLL_RST : entity work.qdr_dll
	port map (
		rst         	=> rst,
		clk200   		=> clk200,
		qdriip_clk		=> qdriip_clk,
		dll_rst_start   => dll_rst_start,
		dll_rst_busy   	=> dll_rst_busy,
		dll_rst_done   	=> dll_rst_done,
		dll_on   		=> dll_on,
		dll_off_n		=> qdriip_dll_off_n
	);

    -- Reset the clock lock solution if system reset is active or the DLL is reset.
	clk_lck_rst <= rst or dll_rst_busy;
	
    -- Optimize timing at the SRAM read port.
    QDR_CLK_LCK : entity work.clock_lock
    generic map (
		SOURCES			=> QDR_CLK_SOURCES,
		MAX_SEARCH		=> QDR_MAX_SEARCH,
		RESET_IF_NOSOL	=> QDR_RESET_IF_NOSOL
	)
	port map (
		rst         	=> clk_lck_rst,
		lock_start   	=> dll_rst_done,
		lock_busy   	=> open,
		lock_status    	=> lock_status,
		clk_lck_clk		=> qdriip_clk,
		clk_fwd			=> cq,
		clk_edge_search	=> QDR_CLK_EDGE,
		clk_edge_detect	=> clk_edge_detect,
		fwd_counter		=> fwd_counter,
		fwd_ld			=> fwd_ld,
		fwd_ce			=> fwd_ce,
		fwd_inc			=> fwd_inc,
		data_ld			=> data_ld,
		data_ce			=> data_ce,
		data_inc		=> data_inc
	);
	
    -- Report the clock lock status.
    clk_lck_status <= lock_status(0);

--************************************************* I/O & IDELAY ASSIGNMENTS ***************************************************

	-- Use double data rate (DDR) output registers to forward a differential write clock to the QDR II+.
	K_ODDR : ODDR
		generic map (DDR_CLK_EDGE => "SAME_EDGE")
		port map (D1 => '1', D2 => '0', C => qdriip_clk, CE => '1', Q => ddr_k, R => rst, S => '0');
    -- Offset the output clock by 180 degrees relative to address and r/w controls.
	K_ODELAY : ODELAYE2
		generic map(ODELAY_TYPE => "FIXED", ODELAY_VALUE => QDR_PERIOD/2, HIGH_PERFORMANCE_MODE => "TRUE")
		port map(C => '0', REGRST => '0', LD => '0', CE => '0', INC => '0', CINVCTRL => '0', CNTVALUEIN => "00000", CLKIN => '0', ODATAIN => ddr_k, LDPIPEEN => '0', DATAOUT =>qdriip_k, CNTVALUEOUT => open);

	-- Use double data rate (DDR) instead of single-ended output registers to supply addresses for consistent timing.
 	SA_BUS : for i in 0 to QDR_ADDR_WIDTH-1 generate
		SA_ODDR : ODDR
			generic map (DDR_CLK_EDGE => "SAME_EDGE")
			port map (D1 => sa(i), D2 => sa(i), C => qdriip_clk, CE => '1', Q => ddr_sa(i), R => rst, S => '0');
		-- Do not offset the address, but use the ODELAY element to match the output clock path.
		SA_ODELAY : ODELAYE2
			generic map(ODELAY_TYPE => "FIXED", ODELAY_VALUE => 0, HIGH_PERFORMANCE_MODE => "TRUE")
			port map(C => '0', REGRST => '0', LD => '0', CE => '0', INC => '0', CINVCTRL => '0', CNTVALUEIN => "00000", CLKIN => '0', ODATAIN => ddr_sa(i), LDPIPEEN => '0', DATAOUT =>qdriip_sa(i), CNTVALUEOUT => open);
	end generate;
	
	-- Use double data rate (DDR) instead of single-ended output registers to supply write enables for consistent timing.
	WREN_ODDR : ODDR
		generic map (DDR_CLK_EDGE => "SAME_EDGE")
		port map (D1 => w_n, D2 => w_n, C => qdriip_clk, CE => '1', Q => ddr_wn, R => rst, S => '0');
	-- Do not offset the write enable, but use the ODELAY element to match the output clock path.
	WREN_ODELAY : ODELAYE2
		generic map(ODELAY_TYPE => "FIXED", ODELAY_VALUE => 0, HIGH_PERFORMANCE_MODE => "TRUE")
		port map(C => '0', REGRST => '0', LD => '0', CE => '0', INC => '0', CINVCTRL => '0', CNTVALUEIN => "00000", CLKIN => '0', ODATAIN => ddr_wn, LDPIPEEN => '0', DATAOUT =>qdriip_w_n, CNTVALUEOUT => open);
	-- Use double data rate (DDR) instead of single-ended output registers to supply read enables for consistent timing.
	RDEN_ODDR : ODDR 
		generic map (DDR_CLK_EDGE => "SAME_EDGE")
		port map (D1 => r_n, D2 => r_n, C => qdriip_clk, CE => '1', Q => ddr_rn, R => rst, S => '0');
	-- Do not offset the read enable, but use the ODELAY element to match the output clock path.
	RDEN_ODELAY : ODELAYE2
		generic map(ODELAY_TYPE => "FIXED", ODELAY_VALUE => 0, HIGH_PERFORMANCE_MODE => "TRUE")
		port map(C => '0', REGRST => '0', LD => '0', CE => '0', INC => '0', CINVCTRL => '0', CNTVALUEIN => "00000", CLKIN => '0', ODATAIN => ddr_rn, LDPIPEEN => '0', DATAOUT =>qdriip_r_n, CNTVALUEOUT => open);
		
	-- Use double data rate (DDR) output registers to supply byte enables on the rising and falling edge of the QDR II+ clock.
 	BWN_BUS : for i in 0 to 1 generate
		BWN_ODDR : ODDR 
			generic map (DDR_CLK_EDGE => "SAME_EDGE") 
			port map (D1 => bw_n(i), D2 => bw_n(i+2), C => qdriip_clk, CE => '1', Q => ddr_bwn(i), R => rst, S => '0');
		-- Offset the byte enables by 90 degrees relative to address and -90 degrees relative to clock.
		BWN_ODELAY : ODELAYE2
			generic map(ODELAY_TYPE => "FIXED", ODELAY_VALUE => QDR_PERIOD/4, HIGH_PERFORMANCE_MODE => "TRUE")
			port map(C => '0', REGRST => '0', LD => '0', CE => '0', INC => '0', CINVCTRL => '0', CNTVALUEIN => "00000", CLKIN => '0', ODATAIN => ddr_bwn(i), LDPIPEEN => '0', DATAOUT =>qdriip_bw_n(i), CNTVALUEOUT => open);
	end generate;

	-- Use double data rate (DDR) output registers to supply write data on the rising and falling edge of the QDR II+ clock.
 	D_BUS : for i in 0 to 17 generate
		D_ODDR : ODDR 
			generic map (DDR_CLK_EDGE => "SAME_EDGE")
			port map (D1 => d(i), D2 => d(i+18), C => qdriip_clk, CE => '1', Q => ddr_d(i), R => rst, S => '0');
		-- Offset the write data by 90 degrees relative to address and -90 degrees relative to clock.
		D_ODELAY : ODELAYE2
			generic map(ODELAY_TYPE => "FIXED", ODELAY_VALUE => QDR_PERIOD/4, HIGH_PERFORMANCE_MODE => "TRUE")
			port map(C => '0', REGRST => '0', LD => '0', CE => '0', INC => '0', CINVCTRL => '0', CNTVALUEIN => "00000", CLKIN => '0', ODATAIN => ddr_d(i), LDPIPEEN => '0', DATAOUT =>qdriip_d(i), CNTVALUEOUT => open);
	end generate;

    -- Set the IDELAY value for the echo clock using the clock lock controller.
	CQ_IDELAY : IDELAYE2
		generic map(IDELAY_TYPE => "VARIABLE", IDELAY_VALUE => QDR_CQ_INIT, HIGH_PERFORMANCE_MODE => "TRUE")
		port map(C => qdriip_clk, REGRST => '0', LD => fwd_ld(0), CE => fwd_ce(0), INC => fwd_inc(0), CINVCTRL => '0', CNTVALUEIN => "00000", IDATAIN => qdriip_cq, DATAIN => '0', LDPIPEEN => '0', DATAOUT =>ddr_cq, CNTVALUEOUT => fwd_counter);
	-- Use a double data rate (DDR) register to capture the state of the echo clock for consistent timing.
	CQ_DDR : IDDR
		generic map(DDR_CLK_EDGE => "SAME_EDGE_PIPELINED")
		port map(Q1 => cq(0), Q2 => open, C => qdriip_clk, CE => '1', D => ddr_cq, R => rst, S => '0');
					
    -- Set the IDELAY value for the read data valid input using the clock lock controller.
	QVLD_IDELAY : IDELAYE2
		generic map(IDELAY_TYPE => "VARIABLE", IDELAY_VALUE => QDR_QVLD_INIT, HIGH_PERFORMANCE_MODE => "TRUE")
		port map(C => qdriip_clk, REGRST => '0', LD => data_ld(0), CE => data_ce(0), INC => data_inc(0), CINVCTRL => '0', CNTVALUEIN => "00000", IDATAIN => qdriip_qvld, DATAIN => '0', LDPIPEEN => '0', DATAOUT =>ddr_qvld, CNTVALUEOUT => open);
		-- Use a double data rate (DDR) input register to capture the valid input on the rising and falling edge of the sample clock.
	QVLD_DDR : IDDR
		generic map(DDR_CLK_EDGE => "SAME_EDGE_PIPELINED")
		port map(Q1 => qvld(0), Q2 => qvld(1), C => qdriip_clk, CE => '1', D => ddr_qvld, R => rst, S => '0');
		
	-- Process each read data input bit.
    Q_BUS : for i in 0 to 17 generate
		-- Set the IDELAY values for each SRAM read data bit using the clock lock controller.
    	Q_IDELAY : IDELAYE2
    		generic map (IDELAY_TYPE  => "VARIABLE", IDELAY_VALUE => QDR_Q_INIT(i), HIGH_PERFORMANCE_MODE => "TRUE")
			port map(C => qdriip_clk, REGRST => '0', LD => data_ld(0), CE => data_ce(0), INC => data_inc(0), CINVCTRL => '0', CNTVALUEIN => "00000", IDATAIN => qdriip_q(i), DATAIN => '0', LDPIPEEN => '0', DATAOUT =>ddr_q(i), CNTVALUEOUT => open);
		-- Use double data rate (DDR) input registers to capture read data on the rising and falling edge of the sample clock.
		Q_DDR : IDDR
			generic map (DDR_CLK_EDGE => "SAME_EDGE_PIPELINED")
			port map (Q1 =>  q(i), Q2 => q(i+18), C => qdriip_clk, CE => '1', D => ddr_q(i), R => rst, S => '0');
	end generate;
	
end behavioral;	