--*******************************************************************************************************************************
--
-- Description: Generic transceiver design applied to all SigFPGA products.
--					
-- Rev History: 08-04-2016:	Added ports (rx_frame_in, tx_frame_out, tx_enable_out) and removed several unused signals to eliminate synthesis warnings.
-- 				02-18-2016:	Changed type of the RX overrange indicator.
-- 				04-17-2015:	Added functionality to program flash through the PCI interface.
--				12-08-2014:	Moved GPIO and USER control bits (direction/tristate) to BAR0 registers.
--				10-20-2014: Fixed the channel number index on the pci_measurement bits passed to the RX and TX datapaths.
--				08-20-2014: Added tx_data_out type, bidirectional SPI data bus (SDIO) support, fan monitor, changed polartiy of gpio_direction and user_direction.
-- 				05-28-2014: Added PCI performance monitor.
--				05-21-2014:	Consolidated QDR II+ SRAM registers and added control signal for the core calibration function.
--				09-04-2013:	Initial release.
--                                                  
--*******************************************************************************************************************************
                                             
library ieee,unisim;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use unisim.VCOMPONENTS.all;
use work.model_specific_package.all;
use work.transceiver_package.all;
use work.dma_package.all;

entity transceiver is
port (
	-- PCI Express bus signals produced by the Xilinx Endpoint Block Plus for PCI Express core generator.
	sys_rst_n			: in std_logic;									-- Refer to the Xilinx PCI Express documentation.
	pci_sys_clk			: in std_logic;									-- Refer to the Xilinx PCI Express documentation.
	pci_exp_rxp			: in std_logic_vector (PCI_LANES-1 downto 0);	-- Refer to the Xilinx PCI Express documentation.
	pci_exp_rxn			: in std_logic_vector (PCI_LANES-1 downto 0);	-- Refer to the Xilinx PCI Express documentation.
	pci_exp_txp			: out std_logic_vector (PCI_LANES-1 downto 0);	-- Refer to the Xilinx PCI Express documentation.
	pci_exp_txn			: out std_logic_vector (PCI_LANES-1 downto 0);	-- Refer to the Xilinx PCI Express documentation.
	
	-- Clocks generated external to the FPGA.
	clk200_in			: in std_logic; 								-- 200 MHz clock from an external oscillator.
	sample_clk_in  		: in std_logic;									-- Sample clock from the external frequency synthesizer or system source.
	
	-- RX data, over-range (ovr) and forwarded clock signals.
	rx_data_in  		: in RX_BUS_IN;									-- RX channel data bus.
	rx_frame_in  		: in RX_FRAME_BIT;								-- RX channel frame sync.
	rx_ovr_in			: in OVR_BIT_IN;								-- RX channel over-range indicator.
	rx_clk_in			: in RX_CLOCK_BIT;								-- RX channel input forwarded clock.
	tx_data_out  		: out TX_BUS_OUT;								-- TX channel data bus.
	tx_frame_out  		: out TX_FRAME_BIT;								-- TX channel frame sync.
	tx_enable_out  		: out TX_ENABLE_BIT;							-- TX channel enable.
	tx_clk_in			: in TX_CLOCK_BIT;								-- TX channel reference clock.
	tx_clk_out			: out TX_CLOCK_BIT;								-- TX channel output forwarded clock.
	
	-- QDR II+ SRAM interface signals.
	qdriip_sys_clk		: in std_logic;									-- QDR II+ SRAM system clock.
	qdriip_cq			: in DUAL_1;									-- QDR II+ SRAM echo clock.
	qdriip_q			: in DUAL_18;									-- QDR II+ SRAM synchronous data outputs from the SRAM.
	qdriip_qvld			: in DUAL_1;									-- QDR II+ SRAM read data valid flag.
	qdriip_k			: out DUAL_1;									-- QDR II+ SRAM input clock.
	qdriip_d			: out DUAL_18;									-- QDR II+ SRAM synchronous data inputs to the SRAM.
	qdriip_sa			: out DUAL_QDR;									-- QDR II+ SRAM synchronous address inputs to the SRAM.
	qdriip_w_n			: out DUAL_1;									-- QDR II+ SRAM synchronous active low write enable.
	qdriip_r_n			: out DUAL_1;									-- QDR II+ SRAM synchronous active low read enable.
	qdriip_bw_n			: out DUAL_2;									-- QDR II+ SRAM synchronous byte enable inputs to the SRAM.
	qdriip_dll_off_n	: out DUAL_1;									-- QDR II+ SRAM active low SRAM DLL disable.
	
	-- Inter-integrated circuit (I2C) bus signals.	
	i2c_scl 			: out std_logic;								-- I2C programming clock.
	i2c_sda_in 			: in std_logic;									-- I2C input data from peripheral chips.
	i2c_sda_out 		: out std_logic;								-- I2C output data to peripheral chips.
	i2c_direction 		: out std_logic;								-- I2C tristate direction control.
	
	-- Serial peripheral interface (SPI) bus signals.
	spi_clk				: out std_logic;								-- SPI programming clock.
	spi_sdi				: out std_logic;								-- SPI data output to peripheral chips.
	spi_sdo				: in  std_logic;								-- SPI data input from peripheral chips.
	spi_direction		: out std_logic;								-- SPI data direction for peripherals with bidirectional bus (SDIO).
	spi_csb_cmd 		: out std_logic_vector(2 downto 0); 			-- SPI chip select command interface (Bit 0: Clock, Bit 1: Data, Bit 2: Enable).
	
	-- Frequency synthesizer command interface signals, applies only to the programmable synthesizer option.
	synth_cmd			: out std_logic_vector(2 downto 0);				-- Dedicated serial command bus to the frequency synthesizer (Bit 0: Clock, Bit 1: Data, Bit 2: Enable).
	
	-- LED control signals.
	led_on				: out std_logic_vector(1 downto 0);				-- Active high control to switch on LEDs.
	
	-- Hardware trigger inputs.
	coax_trig			: in std_logic;									-- Trigger input on coaxial cable.
	gpio_trig			: in std_logic;									-- Trigger input on GPIO cable.
	
	-- Alert notification from the power monitor and command/status port expander.
	alert_n				: in std_logic;									-- Active low status alert.
	
	-- Control and status of the fan attached to the active heat sink on native PCIe products.
	fan_clear			: out std_logic;								-- Active high clear to the FPGA fan monitor on native PCIe.
	fan_status			: in std_logic;									-- Sticky status bit indicating whether fan is active (1) or has failed (0).
	
	-- Uncommitted bidirectional IO pins (GPIO connector and user connector).
	gpio_in				: in std_logic_vector(GPIO_COUNT downto 1);		-- GPIO data input pins.
	gpio_out			: out std_logic_vector(GPIO_COUNT downto 1);	-- GPIO data output pins.
	gpio_direction		: out std_logic_vector(GPIO_COUNT downto 1);	-- Select GPIO direction as input (0) or output (1).
	gpio_tristate		: out std_logic;								-- Tristate all GPIO voltage translators.
	user_in				: in std_logic_vector(USER_COUNT downto 1);		-- User data input pins.
	user_out			: out std_logic_vector(USER_COUNT downto 1);	-- User data output pins.
	user_direction		: out std_logic_vector(USER_COUNT downto 1);	-- Select USER direction as input (0) or output (1).
	user_tristate		: out std_logic;								-- Tristate all USER voltage translators.
	gpio_user_cmd		: out std_logic_vector(2 downto 0);				-- GPIO/user voltage translator command interface (Bit 0: Clock, Bit 1: Data, Bit 2: Load).
	
	-- BPI configuration flash interface bus and controls.
	bpi_a				: out std_logic_vector(22 downto 0);			-- Address bus.
	bpi_rs				: out std_logic_vector(1 downto 0);				-- Revision address.
	bpi_rs_disable		: out std_logic; 								-- Revision select disable.
	bpi_d_in			: in std_logic_vector(15 downto 0);				-- Data input (read) bus.
	bpi_d_out			: out std_logic_vector(15 downto 0);			-- Data output (write) bus.
	bpi_d_tristate		: out std_logic;								-- Active high data bus tristate.
	bpi_adv_b			: out std_logic; 								-- Active low address valid.
	bpi_fcs_b			: out std_logic; 								-- Active low chip enable.
	bpi_fwe_b			: out std_logic; 								-- Active low write enable.
	bpi_foe_b			: out std_logic; 								-- Active low output enable.
	bpi_vpp				: out std_logic 								-- Active high enable voltage for program and erase operations.	
);                                                                                            
end transceiver;

architecture structural of transceiver is

	-- Clock and reset signals.
	signal user_clk			: std_logic;							-- Clock supplied by Xilinx PCI Express Endpoint core (125 MHz or 250 MHz).
	signal user_reset		: std_logic;							-- Active high reset supplied by PCI Express bus.
 	signal rst				: std_logic;							-- Active high reset derived from PCI Express active low reset.
    signal software_reset	: std_logic;							-- Active high reset tied to a register for software control.
     		
	-- Slave (slv) interface connections between the PLDA EZDMA2 core and the application logic.
	signal slv_dataout 			: std_logic_vector(AXI_WIDTH-1 downto 0);		-- Refer to the PLDA EZDMA2 reference manual.
	signal slv_bytevalid 		: std_logic_vector((AXI_WIDTH/8)-1 downto 0);	-- Refer to the PLDA EZDMA2 reference manual.
	signal slv_addr 			: std_logic_vector(63 downto 0);	-- Refer to the PLDA EZDMA2 reference manual.
	signal bar0_write			: std_logic;						-- Logic one indicates that the current PCI write request is directed at BAR0.
	signal bar2_write			: std_logic;						-- Logic one indicates that the current PCI write request is directed at BAR2.
	signal bar0_read			: std_logic;						-- Logic one indicates that the current PCI read request is directed at BAR0.
	signal bar2_read			: std_logic;						-- Logic one indicates that the current PCI read request is directed at BAR2.
	signal bar0_rddata			: std_logic_vector(AXI_WIDTH-1 downto 0);	-- Data return from read operation on BAR0.
	signal bar2_rddata			: std_logic_vector(AXI_WIDTH-1 downto 0);	-- Data return from read operation on BAR2.

	-- Direct memory access (dma) interface connections between the PLDA EZDMA2 core and the application logic.
	signal dma_wr 				: std_logic;						-- Refer to the PLDA EZDMA2 reference manual.
	signal dma_rd 				: std_logic;						-- Refer to the PLDA EZDMA2 reference manual.
	signal dma_rdaddr 			: std_logic_vector(15 downto 0);	-- Refer to the PLDA EZDMA2 reference manual.
	signal dma_wrchannel 		: std_logic_vector(15 downto 0);	-- Refer to the PLDA EZDMA2 reference manual.
	signal dma_rdchannel 		: std_logic_vector(15 downto 0);	-- Refer to the PLDA EZDMA2 reference manual.
	signal dma_regin			: DMA_TYPE_128;						-- Refer to the PLDA EZDMA2 reference manual.
	signal dma_param			: DMA_TYPE_24;						-- Refer to the PLDA EZDMA2 reference manual.
	signal dma_control			: DMA_TYPE_6;						-- Refer to the PLDA EZDMA2 reference manual.
	signal dma_status			: DMA_TYPE_4;						-- Refer to the PLDA EZDMA2 reference manual.
	signal dma_fifocnt			: DMA_TYPE_13;						-- Refer to the PLDA EZDMA2 reference manual.
	signal dma_rddata			: DMA_TYPE_AXI;						-- Refer to the PLDA EZDMA2 reference manual.
	signal dma_wrdata			: std_logic_vector(AXI_WIDTH-1 downto 0);	-- Refer to the PLDA EZDMA2 reference manual.
	signal dma_page_address		: DMA_TYPE_64;						-- Page address stored in the BAR0 memory.
	signal dma_descriptor		: DMA_TYPE_32;						-- Page descriptor stored in the BAR0 memory.
	signal dma_current_page		: DMA_TYPE_10;						-- Page number of the active DMA buffer.
	signal dma_current_burst	: DMA_TYPE_16;						-- Number of the active DMA burst within a page.
	signal dma_marker 			: DMA_TYPE_1;						-- Logic one indicates that the DMA channel has been refreshed.
	signal dma_engine			: DMA_3;							-- Status of DMA engine busy (2), go (1), and enable (0) bits.
	
	-- Serial bus interface signals.
	signal serial_config		: std_logic_vector(9 downto 0);		-- Serial bus command word.
	signal serial_payload		: std_logic_vector(7 downto 0);		-- Serial bus payload (instruction + data).
	signal serial_start			: std_logic;						-- Logic one initiates a SPI bus transaction.
	signal serial_busy			: std_logic;						-- Logic one indicates that the process is active.
	signal serial_shift			: std_logic; 						-- Logic one indicates that the register needs to perform a serial shift.
	signal serial_read			: std_logic; 						-- Registered duplicate of sdo.

	-- BPI configuration flash signals.
	signal bpi_cmd_bus			: std_logic_vector(32 downto 0);	-- Command bus that carries instructions and data to program the BPI flash.
	signal bpi_cmd_full			: std_logic; 						-- Logic one indicates the BPI command FIFO cannot accept another burst.
	signal bpi_reply_bus		: std_logic_vector(32 downto 0); 	-- Reply bus that returns data and errors from the BPI flash.
	signal bpi_reply_rd			: std_logic; 						-- Logic one pops the BPI reply FIFO to the next value.
	
    -- Interrupt controller signals.
	signal int_request			: std_logic;						-- Refer to the PLDA EZDMA2 reference manual.
	signal int_timeout			: std_logic_vector(27 downto 0);	-- Maximum time to hold an interrupt request active.
	signal int_response			: std_logic_vector(27 downto 0);	-- Time taken to service the last interrupt or timeout value.
	signal global_int_mask		: std_logic;						-- Global interrupt mask to enable/disable all interrupts.
	signal global_status		: std_logic_vector(4 downto 0);		-- Consolidated interrupt status bus.  
	signal error_monitor		: std_logic_vector(13 downto 0);	-- Bus used to convey instantaneous error conditions.        
    signal error_status			: std_logic_vector(13 downto 0);	-- Bus used to convey error status since last queried.
	signal error_int_mask		: std_logic_vector(13 downto 0);	-- Mask bits to suppress interrupts based on error status updates.
	signal channel_monitor 		: CH_TYPE_3; 						-- Bus used to convey DMA refresh status and channel errors.
	signal channel_status 		: CH_TYPE_3;  						-- Bus used to convey channel status since last queried.
	signal channel_int_mask 	: CH_TYPE_3;  						-- Mask bits to suppress interrupts based on channel status updates.
	signal fifo_error			: CH_TYPE_1;						-- Bus used to convey FIFO overflow or underlow errors.             
	signal rx_tx_error			: CH_TYPE_1;						-- Bus used to convey ADC or DAC error condition.   
	signal bar0_address_error	: std_logic_vector(1 downto 0);		-- Bus used to convey a BAR0 PCI read or write error.  
	signal bar2_address_error	: std_logic_vector(1 downto 0);		-- Bus used to convey a BAR2 PCI read or write error.     

	-- QDR II+ SRAM control and status signals.
	signal sram_start			: DUAL_1;							-- Logic one for a single period activates the QDR II+ SRAM test.
	signal sram_busy			: DUAL_1;							-- Logic one indicates that the QDR II+ SRAM test is in progress.
	signal sram_error			: DUAL_32;							-- Report the number of errors recorded while testing SRAM-A (0) or SRAM-B (1).
 	signal sram_clk_status		: DUAL_1;							-- Logic one indicates that the QDR II+ SRAM clock is locked.
    signal qdriip_clk			: std_logic;						-- Clock assigned to SRAM-A and SRAM-B.
    signal qdriip_rst			: std_logic;						-- Active high asynchronous SRAM reset.
 	
    -- Data, command and status signals assigned to each channel.
    signal tx_data_in			: TX_BUS;							-- Bus carrying selected data source to the DAC interface core.
    signal rx_data_out			: RX_BUS;							-- Data bus carrying ADC samples from the ADC interface core.
	signal rx_ovr_out			: OVR_BIT;							-- Logic one indicates an over-range condition on the ADC.
	signal data_item			: RX_BUS;							-- Data items that may be stuffed.	
	signal synchronizer_on_off	: CH_TYPE_2;						-- Active high on/off (bit 0/bit 1) switch for the synchronizer.
	signal synchronizer			: CH_TYPE_36;						-- Register containing start and stop event selections.
	signal synchronizer_status	: CH_TYPE_2;						-- Indicator bits for busy (0) and ready (1) status reporting.
	
    -- Clocks, triggers, command and status signals common to multiple channels.
	signal clk200 				: std_logic;						-- 200 MHz clock on global clock buffer.
	signal rx_clk				: std_logic;						-- Clock assigned to the receiver datapath.
	signal tx_clk				: std_logic;						-- Clock assigned to the transmitter datapath.
	signal soft_trig			: std_logic;						-- Single period software trigger generated in the PCI clock domain.
	signal soft_trig_rx			: std_logic;						-- Single period software trigger reclocked in the RX clock domain.
	signal soft_trig_tx			: std_logic;						-- Single period software trigger reclocked in the TX clock domain.
	signal hard_trig_rx			: std_logic;						-- Hardware rising edge start trigger and falling edge stop trigger in the RX clock domain.
	signal hard_trig_tx			: std_logic;						-- Hardware rising edge start trigger and falling edge stop trigger in the TX clock domain.
	signal clk_lck_start		: std_logic;						-- Initiates the clock lock sequence that sets IODELAY values for RX data inputs.
	signal clk_lck_busy			: std_logic;						-- Logic one indicates that the clock lock process is active.
	signal clock_status			: std_logic_vector(7 downto 0);		-- Bus used to report clock stability status.

		-- PCI performance monitor.
 	signal pci_measurement		: std_logic_vector(9 downto 0);		-- Control for PCI throughput and latency measurments.
	signal pci_throughput		: std_logic_vector(63 downto 0);	-- Report the measured PCI DMA throughput in bytes/sec (transmitter=63:32, receiver=31:0).
	signal pci_latency			: std_logic_vector(63 downto 0);	-- Report the maximum PCI DMA latency in 4 ns clock periods (transmitter=63:32, receiver=31:0).
	signal read_completion		: std_logic_vector(1 downto 0);		-- Activity flags for dma_wrchannel(0)and dma_rdchannel(0).
	signal dma_reqchannel		: CH_TYPE_1;						-- Activity flags for dma_wrchannel and dma_rdchannel benchmarks.

begin

--***************************************************** HOST BUS INTERFACE *****************************************************

    -- PCI Express interface including DMA and target transaction decoders.
    PCI : entity work.pci_express
	port map (
		sys_rst_n		=> sys_rst_n,
		pci_sys_clk		=> pci_sys_clk,
		pci_exp_rxp		=> pci_exp_rxp,
		pci_exp_rxn		=> pci_exp_rxn,
		pci_exp_txp		=> pci_exp_txp,
		pci_exp_txn		=> pci_exp_txn,
		user_clk		=> user_clk,
		user_reset		=> user_reset,
		slv_addr 		=> slv_addr,
		slv_dataout 	=> slv_dataout,
		slv_bytevalid 	=> slv_bytevalid,
		bar0_rddata 	=> bar0_rddata,
		bar0_write 		=> bar0_write,
		bar0_read 		=> bar0_read,
		bar2_rddata 	=> bar2_rddata,
		bar2_write 		=> bar2_write,
		bar2_read 		=> bar2_read,
		dma_wr 			=> dma_wr,
		dma_rd 			=> dma_rd,
		dma_rdaddr 		=> dma_rdaddr,
		dma_wrchannel 	=> dma_wrchannel,
		dma_rdchannel 	=> dma_rdchannel,
		dma_regin		=> dma_regin,
		dma_param		=> dma_param,
		dma_control		=> dma_control,
		dma_status		=> dma_status,
		dma_fifocnt		=> dma_fifocnt,
		dma_rddata		=> dma_rddata,
		dma_wrdata		=> dma_wrdata,
		int_request		=> int_request,
		int_timeout		=> int_timeout,
		int_response	=> int_response
	);

    -- Address decoder for base address register (BAR) zero.
    BAR0 : entity work.bar0_decoder
	port map (
	    rst 				=> rst,
		clk 				=> user_clk,
	    slv_addr 			=> slv_addr(13 downto 3),
	    slv_dataout 		=> slv_dataout,
	    slv_bytevalid 		=> slv_bytevalid,
	    bar0_write			=> bar0_write,
	    bar0_read			=> bar0_read,
	    bar0_rddata			=> bar0_rddata,
		dma_rd 				=> dma_rd,
	    dma_rdaddr        	=> dma_rdaddr(13 downto 3),
		dma_rdchannel 		=> dma_rdchannel(0),
		dma_current_page	=> dma_current_page,
		dma_current_burst	=> dma_current_burst,
		dma_page_address	=> dma_page_address,
		dma_descriptor		=> dma_descriptor,
		dma_engine			=> dma_engine,
		global_int_mask 	=> global_int_mask,
		global_status		=> global_status,
		int_timeout			=> int_timeout,
		int_response		=> int_response,
		serial_config		=> serial_config,
		serial_payload		=> serial_payload,
		serial_start		=> serial_start,
		serial_busy			=> serial_busy,
		serial_shift		=> serial_shift,
		serial_read			=> serial_read,
		gpio_direction		=> gpio_direction,
		gpio_tristate		=> gpio_tristate,
		user_direction		=> user_direction,
		user_tristate		=> user_tristate,
		bpi_cmd_bus			=> bpi_cmd_bus,
		bpi_cmd_full		=> bpi_cmd_full,
		bpi_reply_bus		=> bpi_reply_bus,
		bpi_reply_rd		=> bpi_reply_rd,
		clk_lck_start		=> clk_lck_start,
		clk_lck_busy		=> clk_lck_busy,
	    software_reset 		=> software_reset,
		error_monitor 		=> error_monitor,
		error_status 		=> error_status,
		error_int_mask 		=> error_int_mask,
		bar0_address_error	=> bar0_address_error
	);

    -- Address decoder for base address register (BAR) two.
	BAR2 : entity work.bar2_decoder 
	port map (
		rst    			  	=> rst,
		clk		          	=> user_clk,
		bar2_write 		  	=> bar2_write,
	    slv_dataout       	=> slv_dataout,
		slv_bytevalid 	  	=> slv_bytevalid,
	    slv_addr 			=> slv_addr(15 downto 3),
		bar2_read		  	=> bar2_read,
		bar2_rddata		  	=> bar2_rddata,
	    dma_rd            	=> dma_rd,
	    dma_rdaddr        	=> dma_rdaddr(15 downto 3),
		dma_rdchannel     	=> dma_rdchannel(0),
		sram_start			=> sram_start,
		sram_busy			=> sram_busy,
		sram_error			=> sram_error,
		pci_measurement		=> pci_measurement,
		pci_latency			=> pci_latency,
		pci_throughput		=> pci_throughput,
		synchronizer		=> synchronizer,
		synchronizer_on_off	=> synchronizer_on_off,
		synchronizer_status	=> synchronizer_status,
		soft_trig			=> soft_trig,
		led_control			=> led_on,
		channel_monitor 	=> channel_monitor,
		channel_status 		=> channel_status,
		channel_int_mask	=> channel_int_mask,
		bar2_address_error  => bar2_address_error
	);
	
    -- Tie off DMA engine status bits used by BAR0 for unused channels.
	NO_CH : if TX_CHANNEL_COUNT+RX_CHANNEL_COUNT < 8 generate
		NO_DMA : for i in TX_CHANNEL_COUNT+RX_CHANNEL_COUNT+1 to 8 generate
			dma_engine(i) <= (others => '0');
		end generate;
	end generate;

--*********************************************** CLOCK/RESET SIGNAL ASSIGNMENTS ***********************************************

    -- Clock assignments.
    CLK_SYNC : entity work.clock_route
	port map (
		rst				=> rst,
		sample_clk_in	=> sample_clk_in,
		clk200_in		=> clk200_in,
		qdriip_sys_clk	=> qdriip_sys_clk,
		rx_clk 			=> rx_clk,
		tx_clk			=> tx_clk,
		clk200			=> clk200,
		qdriip_clk		=> qdriip_clk,
		qdriip_rst		=> qdriip_rst
	);
        
    -- Generate a global reset from the PCI Express reset or software reset.
	rst <= user_reset or software_reset;
	
--******************************************************** RX CHANNELS *********************************************************

	-- Instantiate the requested number of receiver channels.
	RX_CH : if RX_CHANNEL_COUNT > 0 generate
		-- Interface to the ADC chips that source data to each receiver channel.  
		ADC : entity work.adc_interface
		port map (
			rst         	=> rst,
			rx_clk			=> rx_clk,
			rx_data_out		=> rx_data_out,
			rx_ovr_out		=> rx_ovr_out,
			clk_lck_start	=> clk_lck_start,
			clk_lck_busy	=> clk_lck_busy,
			clk_lck_status  => clock_status(ADC_CLK_SOURCES+3 downto 4),
			rx_data_in		=> rx_data_in,
			rx_frame_in		=> rx_frame_in,
			rx_ovr_in		=> rx_ovr_in,
			rx_clk_in  		=> rx_clk_in
		);
		-- Reclock the hardware trigger in the RX clock domain.
		process (rx_clk)
		begin
			if rising_edge(rx_clk) then
        		-- Allow hardware triggering from either the coax or GPIO trigger inputs.
        		hard_trig_rx <= coax_trig or gpio_trig;
			end if;
		end process;
        -- Reclock the single clock period software trigger in the RX clock domain.
		RX_TRIG : entity work.clock_crossing
		port map (
			rst					=> rst,
			source_clk			=> user_clk,
			target_clk  		=> rx_clk,
			source_to_target	=> soft_trig,
			target_enable		=> soft_trig_rx
		);
		-- Generate a datapath for the number of receiver channels requested.
		RX : for i in 1 to zero_to_one(RX_CHANNEL_COUNT) generate
			-- Instantiate each receiver datapath.
			RX_DATA : entity work.rx_datapath
			port map (
				rst					=> rst,
				user_clk			=> user_clk,
				rx_clk				=> rx_clk,
				data_item			=> rx_data_out(i),
				synchronizer_on_off	=> synchronizer_on_off(i),
				synchronizer_status	=> synchronizer_status(i),
				start_event			=> synchronizer(i)(1 downto 0),
				stop_event			=> synchronizer(i)(3 downto 2),
				sample_count		=> synchronizer(i)(35 downto 4),
				soft_trigger		=> soft_trig_rx,
				hard_trigger		=> hard_trig_rx,
				pci_measurement		=> pci_measurement(i-1),
				dma_page_address	=> dma_page_address(i),
				dma_descriptor		=> dma_descriptor(i),
				dma_current_page	=> dma_current_page(i),
				dma_current_burst	=> dma_current_burst(i),
				dma_rddata			=> dma_rddata(i),
				dma_marker			=> dma_marker(i),
				fifo_error			=> fifo_error(i),
				dma_wr				=> dma_wr,
				dma_wrchannel		=> dma_wrchannel(i),
				dma_rd				=> dma_rd,
				dma_rdchannel		=> dma_rdchannel(i),
				dma_regin			=> dma_regin(i),
				dma_param			=> dma_param(i),
				dma_control			=> dma_control(i),
				dma_status			=> dma_status(i)(3),
				dma_engine			=> dma_engine(i),
				dma_fifocnt			=> dma_fifocnt(i)
			);
			-- Assign over-range bits to the error register assigned to the datapath.
			rx_tx_error(i) <= '0' when to_integer(unsigned(rx_ovr_out(i))) = 0 else '1';
		end generate;
		
		-- Assign unused clock status bits.
		NO_CLK_STAT : if ADC_CLK_SOURCES < 4 generate
		clock_status(7 downto ADC_CLK_SOURCES+4) <= (others => '1');
		end generate;
		
	end generate;

	-- Assign unused signals if there are no RX channels.
	NO_RX : if RX_CHANNEL_COUNT = 0 generate
		tx_clk_out(RX_CHANNEL_COUNT+1) <= '0';
		clock_status(7 downto 4) <= (others => '1');
	end generate;
	
	-- Assign unused rddata inputs associated with TX channels.
	TX_RD : for i in RX_CHANNEL_COUNT+1 to TX_CHANNEL_COUNT+RX_CHANNEL_COUNT generate
		dma_rddata(i) <= (others => '0');
	end generate;
	
--******************************************************** TX CHANNELS *********************************************************

	-- Instantiate the requested number of transmitter channels.
	TX_CH : if TX_CHANNEL_COUNT > 0 generate
		-- Interface to the DAC chips that consume data from each transmitter channel.  
		DAC : entity work.dac_interface
		port map (
			rst         	=> rst,
			tx_clk			=> tx_clk,
			tx_data_in		=> tx_data_in,
			tx_frame_in		=> (others => '0'),
			tx_data_out		=> tx_data_out,
			tx_frame_out	=> tx_frame_out,
			tx_clk_out		=> tx_clk_out
		);
		-- Reclock the hardware trigger in the RX clock domain.
		process (tx_clk)
		begin
			if rising_edge (tx_clk) then
        		-- Allow hardware triggering from either the coax or GPIO trigger inputs.
        		hard_trig_tx <= coax_trig or gpio_trig;
			end if;
		end process;
        -- Reclock the single clock period software trigger in the TX clock domain.
		TX_TRIG : entity work.clock_crossing
		port map (
			rst					=> rst,
			source_clk			=> user_clk,
			target_clk  		=> tx_clk,
			source_to_target	=> soft_trig,
			target_enable		=> soft_trig_tx
		);
		-- Generate a datapath for the number of transmitter channels requested.
		TX : for i in RX_CHANNEL_COUNT+1 to TX_CHANNEL_COUNT+RX_CHANNEL_COUNT generate
			-- Instantiate each transmitter datapath.
			TX_DATA : entity work.tx_datapath
			port map (
				rst					=> rst,
				user_clk			=> user_clk,
				tx_clk				=> tx_clk,
				data_item			=> tx_data_in(i),
				synchronizer_on_off	=> synchronizer_on_off(i),
				synchronizer_status	=> synchronizer_status(i),
				start_event			=> synchronizer(i)(1 downto 0),
				stop_event			=> synchronizer(i)(3 downto 2),
				sample_count		=> synchronizer(i)(35 downto 4),
				soft_trigger		=> soft_trig_tx,
				hard_trigger		=> hard_trig_tx,
				pci_measurement		=> pci_measurement(i-1),
				dma_page_address	=> dma_page_address(i),
				dma_descriptor		=> dma_descriptor(i),
				dma_current_page	=> dma_current_page(i),
				dma_current_burst	=> dma_current_burst(i),
				dma_wrdata			=> dma_wrdata,
				dma_marker			=> dma_marker(i),
				fifo_error			=> fifo_error(i),
				dma_wr				=> dma_wr,
				dma_wrchannel		=> dma_wrchannel(i),
				dma_rd				=> dma_rd,
				dma_rdchannel		=> dma_rdchannel(i),
				dma_regin			=> dma_regin(i),
				dma_param			=> dma_param(i),
				dma_control			=> dma_control(i),
				dma_status			=> dma_status(i)(3),
				dma_engine			=> dma_engine(i),
				dma_fifocnt			=> dma_fifocnt(i)
			);
			-- Assign over-range bits to the error register assigned to the datapath.
			rx_tx_error(i) <= '0';		
		end generate;
	end generate;
	
	-- Assign unused signals if there are no TX channels.
	NO_TX : if TX_CHANNEL_COUNT = 0 generate
		tx_data_out(RX_CHANNEL_COUNT+1) <= (others => '0');
		tx_clk_out(RX_CHANNEL_COUNT+1) <= '0';
		tx_frame_out <= (others => '0');
	end generate;
	
	-- The TX enable is tied active in this demonstration.
	tx_enable_out <= (others => '1');
	
--******************************************************** QDR II+ SRAM ********************************************************

	-- Instantiate the QDR II+ SRAM demonstration if external memory is included on the product.
	SRAM : if QDR_ADDR_WIDTH > 0 generate
		-- Assign one instance to SRAM-A (0) and another to SRAM-B (1).
		SRAM_X2 : for i in 0 to 1 generate
			QDRII_SRAM : entity work.qdr_demo 
			port map (
				rst					=> qdriip_rst,
				clk					=> user_clk,
				clk200  			=> clk200,
				qdriip_clk   		=> qdriip_clk,
				sram_start			=> sram_start(i),
				sram_busy			=> sram_busy(i),
				sram_error			=> sram_error(i),
				dll_rst_start		=> clock_status(0),
				clk_lck_status		=> sram_clk_status(i),
				qdriip_cq			=> qdriip_cq(i),
				qdriip_q			=> qdriip_q(i),
				qdriip_qvld			=> qdriip_qvld(i),
				qdriip_k			=> qdriip_k(i),
				qdriip_d			=> qdriip_d(i),
				qdriip_sa			=> qdriip_sa(i),
				qdriip_w_n			=> qdriip_w_n(i),
				qdriip_r_n			=> qdriip_r_n(i),
				qdriip_bw_n			=> qdriip_bw_n(i),
				qdriip_dll_off_n	=> qdriip_dll_off_n(i)
			);
			end generate;
		clock_status(3 downto 2) <= sram_clk_status(1) & sram_clk_status(0);
	end generate;
			
	-- Assign unused signals if there is no QDR II+ SRAM.
	NO_SRAM : if QDR_ADDR_WIDTH = 0 generate
		NO_SRAM_X2 : for i in 0 to 1 generate
			sram_busy(i) <= '0';
			sram_error(i) <= (others => '0');
			qdriip_k(i) <= '0';
			qdriip_d(i) <= (others => '0');
			qdriip_sa(i) <= (others => '0');
			qdriip_w_n(i) <= '0';
			qdriip_r_n(i) <= '0';
			qdriip_bw_n(i) <= (others => '0');
			qdriip_dll_off_n(i) <= '0';
		end generate;
		clock_status(3 downto 2) <= (others => '1');
	end generate;
	
--******************************************************** CONTROLLERS *********************************************************

    -- Serial bus controller.
    SERIAL : entity work.serial_bus
	port map (		
		rst    				=> rst,
		clk					=> user_clk,
		serial_payload		=> serial_payload,
		serial_start		=> serial_start,
		serial_busy			=> serial_busy,
		serial_shift		=> serial_shift,
		serial_read			=> serial_read,
		payload_size		=> serial_config(2 downto 0),
		instruction_size	=> serial_config(9 downto 7),
		multi_function		=> serial_config(3),
		bus_select			=> serial_config(6 downto 4),
		i2c_scl 			=> i2c_scl,
		i2c_sda_in 			=> i2c_sda_in,
		i2c_sda_out 		=> i2c_sda_out,
		i2c_direction 		=> i2c_direction,
	    spi_csb_cmd			=> spi_csb_cmd,
	    spi_clk				=> spi_clk,
	    spi_sdi				=> spi_sdi, 
	    spi_sdo				=> spi_sdo,
	    spi_direction		=> spi_direction,
		synth_cmd			=> synth_cmd,
		gpio_user_cmd		=> gpio_user_cmd
	);

    -- BPI configuration flash controller.
    BPI : entity work.bpi_flash
	port map (		
		rst 				=> rst,
		user_clk			=> user_clk,
		clk200				=> clk200,
		bpi_cmd_bus			=> bpi_cmd_bus,
		bpi_cmd_full		=> bpi_cmd_full,
		bpi_reply_bus		=> bpi_reply_bus,
		bpi_reply_rd		=> bpi_reply_rd,
		clk20_locked		=> clock_status(1),
		bpi_a				=> bpi_a,
		bpi_rs				=> bpi_rs,
	    bpi_rs_disable		=> bpi_rs_disable,
	    bpi_d_in			=> bpi_d_in,
	    bpi_d_out			=> bpi_d_out, 
	    bpi_d_tristate		=> bpi_d_tristate,
	    bpi_adv_b			=> bpi_adv_b,
		bpi_fcs_b			=> bpi_fcs_b,
		bpi_fwe_b			=> bpi_fwe_b,
		bpi_foe_b			=> bpi_foe_b,
		bpi_vpp				=> bpi_vpp
	);
	
    -- Controller to generate optional interrupts triggered by various error and status conditions.
	INTERRUPT : entity work.interrupt_controller
	port map (
		rst 				=> rst,
		clk		        	=> user_clk,
		rx_clk 				=> rx_clk,
		tx_clk 				=> tx_clk,
		int_request 		=> int_request,
		error_status 		=> error_status,
		error_int_mask 		=> error_int_mask,
		error_monitor 		=> error_monitor,
		channel_status 		=> channel_status,
		channel_int_mask 	=> channel_int_mask,
		channel_monitor 	=> channel_monitor,
		bar0_address_error 	=> bar0_address_error,
		bar2_address_error 	=> bar2_address_error,
		fifo_error 			=> fifo_error,
		rx_tx_error 		=> rx_tx_error,
		clock_status		=> clock_status,
		alert_n				=> alert_n,
		fan_status			=> fan_status,
		dma_marker 			=> dma_marker,
		global_int_mask 	=> global_int_mask,
		global_status		=> global_status
	);

	-- Clear the fan monitor during a global reset.
	fan_clear <= rst;
    
	-- PCI throughput and latency measurements.
	read_completion <= dma_wrchannel(0) & dma_rdchannel(0);
	RX_PERF : if RX_CHANNEL_COUNT > 0 generate
		RX_REQ : for i in 1 to RX_CHANNEL_COUNT generate
			dma_reqchannel(i) <= dma_rdchannel(i);
		end generate;
	end generate;
	TX_PERF : if TX_CHANNEL_COUNT > 0 generate
		TX_REQ : for i in RX_CHANNEL_COUNT+1 to TX_CHANNEL_COUNT+RX_CHANNEL_COUNT generate
			dma_reqchannel(i) <= dma_wrchannel(i);
		end generate;
	end generate;
	PCI_PERF : entity work.pci_performance
	port map (		
		rst 			=> rst,
		clk				=> user_clk,
		dma_wr			=> dma_wr,
		dma_rd			=> dma_rd,
		read_completion	=> read_completion,
		dma_reqchannel	=> dma_reqchannel,
		dma_control		=> dma_control,
		pci_measurement	=> pci_measurement(9 downto 8),
		pci_throughput	=> pci_throughput,
		pci_latency		=> pci_latency
	);

	--  Instantiate the IDELAYCTRL primitive if the product includes RX channels or QDR SRAM.
	IDELAY_CTRL : if RX_CHANNEL_COUNT > 0 or QDR_ADDR_WIDTH > 0 generate
		XCVR_IDELAY : IDELAYCTRL
			port map (RDY => clock_status(0), REFCLK => clk200, RST => rst);	       
	end generate;

--************************************************ GPIO/USER SIGNAL ASSIGNMENTS ************************************************
	
    -- The GPIO port is not used in this demostration, so the output signals are tied off below.
	gpio_out <= (others => '0');
	
    -- The user port is not used in this demostration, so the output signals are tied off below.
	user_out <= (others => '0');
	
end structural;